/* Copyright (c) 2016, NVIDIA CORPORATION.  All rights reserved.
 *
 * This software is licensed under the terms of the GNU General Public
 * License version 2, as published by the Free Software Foundation, and
 * may be copied, distributed, and modified under those terms.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#include <linux/delay.h>
#include "nvi.h"
#include "nvi_dmp_icm.h"

#define DMP_FREQ			(102)
#define DMP_PERIOD_US			(9804)
#define PEDOMETER_FREQ			(DMP_FREQ >> 1)
#define DEFAULT_ACCEL_GAIN		33554432L
#define BASE_SAMPLE_RATE		(1125)
#define DMP_DIVIDER			(BASE_SAMPLE_RATE / DMP_FREQ)


static int nvi_mem_rdwr(struct nvi_state *st, u16 addr, u16 len,
			u32 val, u32 msk)
{
	u8 buf_rd[4];
	u32 buf_wr;
	unsigned int i;
	int ret;

	ret = nvi_mem_rd(st, addr, len, buf_rd);
	if (!ret) {
		buf_wr = 0;
		/* convert to little endian */
		for (i = 0; i < len; i++) {
			buf_wr <<= 8;
			buf_wr |= buf_rd[i];
		}

		buf_wr &= ~msk;
		buf_wr |= val;
		ret = nvi_mem_wr_be(st, addr, len, buf_wr);
	}
	return ret;
}

static int nvi_dmp_cntl(struct nvi_state *st, bool en, u16 addr, u32 msk)
{
	u32 val = 0;

	if (en)
		val = msk;
	return nvi_mem_rdwr(st, addr, 2, val, msk);
}

static int nvi_dmp_press_init(struct nvi_state *st)
{
	return 1;
}

static int nvi_dmp_cpass_init(struct nvi_state *st)
{
	int ret = 1;

	if (st->aux.port[0].nmp.type == SECONDARY_SLAVE_TYPE_COMPASS) {
		if (st->snsr[DEV_AUX].enable & (1 << 0))
			ret = nvi_aux_delay(st, __func__);
	}
	return ret;
}

static int nvi_dmp_sm_init(struct nvi_state *st)
{
	int ret;

	ret = nvi_mem_wr_be(st, SMD_MOT_THLD, 4,
			    st->snsr[DEV_SM].cfg.thresh_lo << 16);
	ret |= nvi_mem_wr_be(st, SMD_DELAY_THLD, 4,
			     st->snsr[DEV_SM].cfg.thresh_hi);
	ret |= nvi_mem_wr_be(st, SMD_DELAY2_THLD, 4,
			     st->snsr[DEV_SM].cfg.delay_us_max);
	return ret;
}

static struct nvi_dmp_dev nvi_dmp_devs[] = {
	{
		.dev			= DEV_ACC,
		.data_n			= 12,
		.hdr_n			= 2,
		.hdr			= { 0x80, 0x00, 0x00, 0x00 },
		.hdr_msk		= { 0xFF, 0xF0, 0x00, 0x00 },
		.icm			= {
			.en_addr	= DATA_OUT_CTL1,
			.en_msk		= 0x8000,
			.en_msk		= 0,
			.odr_cfg	= ODR_ACCEL,
			.odr_cntr	= ODR_CNTR_ACCEL,
		},
	},
	{
		.dev			= DEV_GYR,
		.data_n			= 6,
		.hdr_n			= 2,
		.hdr			= { 0x40, 0x00 },
		.hdr_msk		= { 0xFF, 0xF0 },
		.icm			= {
			.en_addr	= DATA_OUT_CTL1,
			.en_msk		= 0x4000,
			.odr_cfg	= ODR_GYRO,
			.odr_cntr	= ODR_CNTR_GYRO,
		},
	},
	{
		.dev			= DEV_QTN,
		.data_n			= 12,
		.hdr_n			= 4,
		.hdr			= { 0x08, 0x00, 0x04, 0x00 },
		.hdr_msk		= { 0xFF, 0xF0, 0xFF, 0xF0 },
		.icm			= {
			.en_addr	= DATA_OUT_CTL1,
			.en_msk		= 0x0800,
			.en_msk		= 0,
			.odr_cfg	= ODR_QUAT6,
			.odr_cntr	= ODR_CNTR_QUAT6,
		},
	},
	{
		.dev			= DEV_SM,
		.data_n			= 1,
		.hdr_n			= 0,
		.fn_init		= &nvi_dmp_sm_init,
		.icm			= {
			.en_addr	= DATA_OUT_CTL1,
			.en_msk		= 0,
			.odr_cfg	= 0,
			.odr_cntr	= 0,
		},
	},
	{
		.dev			= DEV_AUX,
		.data_n			= 6,
		.aux_port		= 0,
		.hdr_n			= 2,
		.hdr			= { 0x20, 0x00 },
		.hdr_msk		= { 0xFF, 0xF0 },
		.fn_init		= &nvi_dmp_cpass_init,
		.icm			= {
			.en_addr	= DATA_OUT_CTL1,
			.en_msk		= 0x2000,
			.odr_cfg	= ODR_CPASS,
			.odr_cntr	= ODR_CNTR_CPASS,
		},
	},
	{
		.dev			= DEV_AUX,
		.data_n			= 6,
		.aux_port		= 2,
		.hdr_n			= 2,
		.hdr			= { 0x80, 0x00 },
		.hdr_msk		= { 0xFF, 0xF0 },
		.fn_init		= &nvi_dmp_press_init,
		.icm			= {
			.en_addr	= DATA_OUT_CTL1,
			.en_msk		= 0x0080,
			.odr_cfg	= ODR_PRESSURE,
			.odr_cntr	= ODR_CNTR_PRESSURE,
		},
	},
};

static int nvi_dd_odr(struct nvi_state *st, struct nvi_dmp_dev *dd)
{
	u16 odr_cfg;
	int ret;

	if (dd->dev == DEV_AUX)
		odr_cfg = (st->aux.port[dd->aux_port].period_us /
			   DMP_PERIOD_US) - 1;
	else
		odr_cfg = (st->snsr[dd->dev].period_us / DMP_PERIOD_US) - 1;
	ret = nvi_mem_wr_be(st, dd->icm.odr_cntr, 2, 0);
	ret |= nvi_mem_wr_be(st, dd->icm.odr_cfg, 2, odr_cfg);
	ret |= nvi_mem_wr_be(st, dd->icm.odr_cntr, 2, 0);
	return ret;
}

static int nvi_dd_en(struct nvi_state *st, struct nvi_dmp_dev *dd)
{
	int ret;

	if (st->snsr[dd->dev].enable) {
		if (dd->fn_init)
			ret = dd->fn_init(st);
		else
			ret = 0;
		if (ret > 0) {
			/* ignore without error */
			ret = 0;
		} else if (!ret) {
			ret = nvi_dd_odr(st, dd);
			ret |= nvi_dmp_cntl(st, true, dd->icm.en_addr,
					    dd->icm.en_msk);
		}
		if (ret && (dd->dev != DEV_AUX))
			st->en_msk &= ~(1 << dd->dev);
	} else {
		ret = nvi_dmp_cntl(st, true, dd->icm.en_addr, dd->icm.en_msk);
	}
	return ret;
}

static struct nvi_dmp_dev *nvi_dd(struct nvi_state *st,
				  unsigned int dev, int port)
{
	unsigned int i;

	for (i = 0; i < ARRAY_SIZE(nvi_dmp_devs); i++) {
		if (dev == nvi_dmp_devs[i].dev) {
			if (dev == DEV_AUX) {
				if (port != nvi_dmp_devs[i].aux_port)
					continue;
			}
			return &nvi_dmp_devs[i];
		}
	}

	return NULL;
}

static int nvi_dd_batch(struct nvi_state *st, unsigned int dev, int port)
{
	struct nvi_dmp_dev *dd;

	dd = nvi_dd(st, dev, port);
	if (dd == NULL)
		return -EINVAL;

	return nvi_dd_odr(st, dd);
}

static int nvi_dd_enable(struct nvi_state *st, unsigned int dev, int port)
{
	struct nvi_dmp_dev *dd;

	dd = nvi_dd(st, dev, port);
	if (dd == NULL)
		return -EINVAL;

	return nvi_dd_en(st, dd);
}

static int nvi_dd_init(struct nvi_state *st, unsigned int dev)
{
	struct nvi_dmp_dev *dd;
	int ret = -EINVAL;

	dd = nvi_dd(st, dev, -1);
	if (dd == NULL)
		return -EINVAL;

	if (dd->fn_init)
		ret = dd->fn_init(st);
	return ret;
}

#define PED_ACCEL_GAIN           67108864L

struct inv_accel_cal_params {
	unsigned int freq;
	unsigned int rate;
	unsigned int gain;
	unsigned int alpha;
	unsigned int a;
};

static struct inv_accel_cal_params accel_cal_para[] = {
	{
		.freq = 1000,
	},
	{
		.freq = 225,
		.rate = 3,
		.gain = DEFAULT_ACCEL_GAIN,
		.alpha = 858993459,
		.a     = 214748365,
	},
	{
		.freq = 102,
		.gain = DEFAULT_ACCEL_GAIN,
		.rate = 1,
		.alpha = 858993459,
		.a     = 214748365,
	},
	{
		.freq = PEDOMETER_FREQ,
		.gain = PED_ACCEL_GAIN,
		.rate = 0,
		.alpha = 858993459,
		.a     = 214748365,
	},
	{
		.freq = 15,
		.gain = DEFAULT_ACCEL_GAIN,
		.rate = 0,
		.alpha = 357913941,
		.a     = 715827883,
	},
	{
		.freq = 5,
		.gain = DEFAULT_ACCEL_GAIN,
		.rate = 0,
		.alpha = 107374182,
		.a     = 966367642,
	},
};

static int nvi_dmp_en(struct nvi_state *st)
{
	unsigned int i;
	int ret = 0;



	st->sts ^= NVI_DBG_SPEW_MSG;
	st->rc_dis = true;


	nvi_wr_pm1(st, __func__, 0x01);
	nvi_i2c_wr_rc(st, &st->hal->reg->pm2, 0x00, __func__, &st->rc.pm2);
	nvi_i2c_wr_rc(st, &st->hal->reg->fifo_cfg, 0x01,
		      __func__, &st->rc.fifo_cfg);
	nvi_i2c_wr_rc(st, &st->hal->reg->fifo_sz, 0x03,
		      __func__, &st->rc.fifo_sz);


	ret |= nvi_i2c_write_rc(st, &st->hal->reg->fifo_en, 0,
				__func__, (u8 *)&st->rc.fifo_en, false);


	nvi_i2c_wr(st, &st->hal->reg->fifo_rst, 0x1F, __func__);
	nvi_i2c_wr(st, &st->hal->reg->fifo_rst, 0x1E, __func__);


	/* reset DMP and FIFO */
	nvi_i2c_wr_rc(st, &st->hal->reg->user_ctrl, 0x0C,
		      __func__, &st->rc.user_ctrl);

	msleep(20);

	for (i = 0; i < ARRAY_SIZE(nvi_dmp_devs); i++)
		ret |= nvi_dd_en(st, &nvi_dmp_devs[i]);


	ret |= nvi_mem_wr_be(st, DATA_INTR_CTL, 2, 0xC000);

	ret |= nvi_mem_wr_be(st, ACCEL_CAL_RATE, 2, accel_cal_para[i].rate);
	ret |= nvi_mem_wr_be(st, PED_RATE, 4, accel_cal_para[i].rate);
	ret |= nvi_mem_wr_be(st, ACCEL_ALPHA_VAR, 4, accel_cal_para[i].alpha);
	ret |= nvi_mem_wr_be(st, ACCEL_A_VAR, 4, accel_cal_para[i].a);
	ret |= nvi_mem_wr_be(st, ACCEL_ONLY_GAIN, 4, accel_cal_para[i].gain);
	/* inv_enable_accel_cal_V3 */
	ret |= nvi_dmp_cntl(st, st->en_msk & (1 << DEV_ACC),
			    MOTION_EVENT_CTL, ACCEL_CAL_EN << 8);
	/* inv_enable_gyro_cal_V3 */
	ret |= nvi_dmp_cntl(st, st->en_msk & (1 << DEV_GYR),
			    MOTION_EVENT_CTL, GYRO_CAL_EN << 8);
	/* inv_enable_compass_cal_V3 */
	ret |= nvi_dmp_cntl(st, 0, MOTION_EVENT_CTL, COMPASS_CAL_EN); /* dis */
	/* inv_enable_9axes_V3 */
	ret |= nvi_dmp_cntl(st, 1, MOTION_EVENT_CTL, NINE_AXIS_EN); /* en */
	/* inv_setup_events */
	ret |= nvi_dmp_cntl(st, st->en_msk & (1 << DEV_STP),
			    MOTION_EVENT_CTL, PEDOMETER_EN << 8);
	ret |= nvi_dmp_cntl(st, st->en_msk & (1 << DEV_SM),
			    MOTION_EVENT_CTL, SMD_EN << 8);
	/* inv_set_wom */
	ret |= nvi_mem_wr_be(st, WOM_ENABLE, 4,
			     !!(st->en_msk & (1 << DEV_SM)));


	st->en_msk |= (1 << DEV_DMP);
	nvi_int_able(st, __func__, true); /* 0x38 = 0x02 */

	nvi_i2c_wr_rc(st, &st->hal->reg->user_ctrl, 0xC0,
		      __func__, &st->rc.user_ctrl);
	return 0;
}

static int nvi_dmp_init(struct nvi_state *st)
{
	u8 init[] = { 0, 0 };
	u8 val;
	s8 pll;
	s32 gryo_sf;
	int a;
	int r;
	int t;
	int ret;

	/* initialize DMP control registers */
	ret = nvi_mem_wr(st, DATA_OUT_CTL1, sizeof(init), init, false);
	ret |= nvi_mem_wr(st, DATA_OUT_CTL2, sizeof(init), init, false);
	ret |= nvi_mem_wr(st, DATA_INTR_CTL, sizeof(init), init, false);
	ret |= nvi_mem_wr(st, MOTION_EVENT_CTL, sizeof(init), init, false);
	if (ret)
		return ret;

	/* initialize GYRO_SF */
	ret = nvi_i2c_rd(st, &st->hal->reg->tbc_pll, &val);
	pll = abs(val & 0x7F);
	if (val & 0x80)
		pll = -pll;
	t = 102870L + 81L * pll;
	a = (1L << 30) / t;
	r = (1L << 30) - a * t;
	gryo_sf = a * 797 * DMP_DIVIDER;
	gryo_sf += (s64)((a * 1011387LL * DMP_DIVIDER) >> 20);
	gryo_sf += r * 797L * DMP_DIVIDER / t;
	gryo_sf += (s32)((s64)((r * 1011387LL * DMP_DIVIDER) >> 20)) / t;
	gryo_sf <<= 1;
	ret |= nvi_mem_wr_be(st, GYRO_SF, sizeof(gryo_sf), gryo_sf);
	if (ret)
		return ret;

	return ret;
}

static const u8 const dmp_fw_20628[] = {
	/* bank 0: 0x0090 */
	0x00, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x05, 0x00, 0x00, 0x00, 0x05, 0x00, 0x00,
	0x00, 0x05, 0x00, 0x02, 0x00, 0x05, 0x00, 0xFF,
	0xFF, 0xF7, 0x00, 0x05, 0x00, 0x05, 0x00, 0x05,
	0x00, 0x05, 0x00, 0x05, 0x00, 0x05, 0x00, 0x05,
	0x80, 0x00, 0x80, 0x00, 0x40, 0x00, 0x40, 0x00,
	0x20, 0x00, 0x20, 0x00, 0x10, 0x00, 0x10, 0x00,
	0x08, 0x00, 0x08, 0x00, 0x04, 0x00, 0x04, 0x00,
	0x02, 0x00, 0x02, 0x00, 0x01, 0x00, 0x01, 0x00,
	0x00, 0x80, 0x00, 0x80, 0x00, 0x40, 0x00, 0x40,
	0x00, 0x20, 0x00, 0x20, 0x00, 0x10, 0x00, 0x10,
	0x00, 0x08, 0x00, 0x08, 0x00, 0x04, 0x00, 0x04,
	0x00, 0x02, 0x00, 0x02, 0x00, 0x01, 0x00, 0x01,
	/* bank 1: 0x0100 */
	0x00, 0x00, 0x0A, 0x8C, 0x00, 0x01, 0xD4, 0xC0,
	0x00, 0x00, 0x00, 0x00, 0x04, 0x00, 0x00, 0x00,
	0x00, 0x08, 0x94, 0xA5, 0x00, 0x02, 0xB6, 0xA3,
	0x13, 0x5C, 0x28, 0xF6, 0x0C, 0xF5, 0xC2, 0x8F,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0xFF, 0xF8, 0x00, 0x38,
	0x04, 0xF6, 0xE8, 0xF4, 0x00, 0x00, 0x68, 0x00,
	0x00, 0x01, 0xFF, 0xC7, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x01, 0x47, 0xAE, 0x14,
	0x3E, 0xB8, 0x51, 0xEC, 0x00, 0x0F, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x40, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x40, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x40, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x01, 0x00, 0x00, 0x01, 0x0C, 0x00,
	0x00, 0x00, 0x01, 0x5E, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0xD6, 0x2B, 0x80,
	0x38, 0x00, 0x00, 0x00, 0x30, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xCC,
	0xFF, 0xFF, 0x00, 0x01, 0x00, 0x06, 0x00, 0x01,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x50, 0x50, 0x50, 0x3F, 0xAF, 0xAF, 0xB0,
	0x00, 0x35, 0xA1, 0x09, 0x36, 0x66, 0x66, 0x66,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x8E, 0x17, 0x00, 0x07, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	/* bank 2: 0x0200 */
	0x40, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x7F, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x07, 0x80,
	0x40, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x3E, 0x03, 0x30, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x02, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x9C, 0x40, 0x0C, 0xCC, 0xCC, 0xCD,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x60, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x3E, 0xB8, 0x51, 0xEC, 0x01, 0x47, 0xAE, 0x14,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x33, 0x33, 0x33, 0x33, 0x0C, 0xCC, 0xCC, 0xCD,
	0x00, 0x00, 0x00, 0x00, 0x01, 0x02, 0x4B, 0xD1,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x52, 0x00, 0x0A,
	0x06, 0xD5, 0x8E, 0x27, 0xE1, 0x05, 0x86, 0xB2,
	0x38, 0xB3, 0x8D, 0x1F, 0xCF, 0x87, 0xC3, 0x8F,
	0x06, 0xD5, 0x8E, 0x27, 0x1F, 0xA1, 0xAB, 0x12,
	0x3A, 0x50, 0xFF, 0xCB, 0x31, 0x7D, 0xD5, 0x43,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	/* bank 3: 0x0300 */
	0x00, 0x00, 0x00, 0x01, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x01, 0x00, 0x00, 0x00, 0x00, 0x80, 0x00,
	0x00, 0x06, 0x36, 0x38, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x11, 0x23, 0xDF,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x07, 0x00, 0x10, 0x00, 0x96, 0x00, 0x3C,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x0D, 0x35, 0xA2, 0x0E, 0xCA, 0x45, 0xB3, 0x53,
	0x53, 0xB6, 0x82, 0x59, 0xC4, 0xDB, 0x1C, 0x13,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x10, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x08,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x10, 0x00, 0x00, 0x00, 0x3D,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x20, 0x31, 0x88, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x03, 0x00, 0x00, 0x00, 0x03,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x0E, 0x49, 0x6E, 0x64, 0xC5, 0x92, 0x24, 0x82,
	0x59, 0xFA, 0xEE, 0x8D, 0xC2, 0x29, 0xFE, 0xD7,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x02,
	0x00, 0x00, 0x00, 0x14, 0x00, 0x28, 0xCC, 0xCC,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	/* bank 4: 0x0400 */
	0x00, 0xC8, 0x00, 0x00, 0x00, 0x00, 0x00, 0x02,
	0x00, 0x00, 0x00, 0x00, 0x03, 0xE8, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x03, 0x3F, 0xC1, 0xA7, 0x68,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x40, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x32, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x40, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	/* bank 5: 0x0500 */
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x10,
	0x00, 0x00, 0x01, 0x00, 0x00, 0x00, 0x10, 0x00,
	0x00, 0x01, 0x00, 0x00, 0x04, 0x00, 0x00, 0x00,
	0x00, 0x40, 0x00, 0x00, 0x00, 0x04, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x0C,
	0x00, 0x00, 0x1F, 0x40, 0x00, 0x80, 0x00, 0x00,
	0x09, 0x24, 0x92, 0x49, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x36, 0xDB, 0x6D, 0xB7, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x40, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01,
	0x3F, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x33, 0x33, 0x33, 0x33, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x0C, 0xCC, 0xCC, 0xCD, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x9D, 0x49, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x01, 0x00, 0x00, 0x00, 0x01, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x07,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	/* bank 6: 0x0600 */
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x01, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x32, 0x00, 0x00, 0x00, 0x32, 0x00, 0x00,
	0x00, 0x32, 0x00, 0x00, 0x00, 0x32, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x32, 0x00, 0x00, 0x00, 0x32, 0x00, 0x00,
	0x00, 0x32, 0x00, 0x00, 0x00, 0x32, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x01, 0x00, 0x00, 0x00, 0x01,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x10, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x13, 0x88, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x10, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x06, 0x40, 0x00, 0x00, 0x0C, 0xCD,
	0x00, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	/* bank 7: 0x0700 */
	0x00, 0x00, 0x50, 0x00, 0x00, 0x00, 0x00, 0x01,
	0x00, 0x00, 0x0E, 0x00, 0x00, 0x00, 0x00, 0x46,
	0x00, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x80, 0x00, 0x00, 0x00, 0x80, 0x00, 0x00,
	0x00, 0x80, 0x00, 0x00, 0x00, 0x80, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x00, 0x01,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x24, 0xF8,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01,
	0x00, 0x80, 0x00, 0x00, 0x00, 0x00, 0x0C, 0xF6,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	/* bank 8: 0x0800 */
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x03, 0x00, 0x00, 0x00, 0xEA,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x7F, 0x00, 0x00, 0x01, 0x4E,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x04, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x01, 0x47, 0xAE, 0x14, 0x4E, 0x40, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x0C, 0x00, 0x00, 0x00, 0xF8,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x01, 0x00, 0x00, 0x4E, 0x40, 0x00, 0x00,
	0x4A, 0x40, 0x00, 0x00, 0x00, 0x00, 0x0D, 0x06,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0xD8, 0xDC, 0xB8, 0xB0, 0xB4, 0xF3, 0xAA, 0xF8,
	0xF9, 0xD1, 0xD9, 0x88, 0x9A, 0xF8, 0xF7, 0x3E,
	0xD8, 0xF3, 0x8A, 0x9A, 0xA7, 0x31, 0xD1, 0xDA,
	0xF1, 0xFF, 0xD8, 0xF1, 0xA4, 0xDF, 0xA5, 0xDE,
	0xF3, 0xA8, 0xDE, 0xD0, 0xDF, 0xA4, 0x84, 0x9F,
	0x24, 0xF2, 0xA9, 0xF8, 0xF9, 0xD1, 0xDA, 0xDE,
	/* bank 9: 0x0900 */
	0xA8, 0xDE, 0xDF, 0xDF, 0xDF, 0xD8, 0xF4, 0xB1,
	0x8D, 0xF3, 0xA8, 0xD0, 0xC0, 0xF1, 0xBC, 0xB0,
	0xBD, 0xBE, 0x82, 0xA1, 0xC3, 0xC5, 0xC7, 0xBC,
	0xBC, 0xA2, 0xDF, 0xDF, 0xDF, 0x80, 0xF2, 0xA2,
	0xC3, 0xCB, 0xC5, 0xBC, 0xF1, 0x90, 0xFC, 0xC3,
	0x00, 0xD9, 0xF4, 0x09, 0x8E, 0xD8, 0xF3, 0xA0,
	0xDF, 0xF1, 0x82, 0x91, 0xA9, 0x2D, 0x55, 0x7D,
	0xA9, 0x80, 0x90, 0xFC, 0x51, 0x00, 0x10, 0xFC,
	0x51, 0x00, 0x10, 0xFC, 0x51, 0x00, 0x10, 0xFC,
	0xC1, 0x04, 0xD9, 0xF2, 0xA0, 0xDF, 0xF4, 0x09,
	0x8E, 0xD8, 0xF6, 0xA0, 0xFA, 0x80, 0x90, 0x38,
	0xF3, 0xDE, 0xDA, 0xF8, 0xF4, 0x09, 0x8E, 0xD8,
	0xF1, 0xBD, 0x95, 0xFC, 0xC1, 0x04, 0xD9, 0xBD,
	0xBD, 0xF4, 0x09, 0x8E, 0xDA, 0xBD, 0xBD, 0xD8,
	0xF6, 0xBC, 0xBC, 0xBD, 0xBD, 0xBE, 0xBE, 0xB5,
	0xA7, 0x84, 0x92, 0x1A, 0xF8, 0xF9, 0xD1, 0xDB,
	0x84, 0x93, 0xF7, 0x6A, 0xB6, 0x87, 0x96, 0xF3,
	0x09, 0xFF, 0xDA, 0xBC, 0xBD, 0xBE, 0xD8, 0xF3,
	0xBC, 0xBC, 0xBD, 0xBD, 0xBE, 0xBE, 0xB0, 0xB4,
	0x8F, 0xF4, 0xB9, 0xAF, 0xD0, 0xC7, 0xBE, 0xBE,
	0xB8, 0xAE, 0xD0, 0xF3, 0x9F, 0x5C, 0xB2, 0x86,
	0xF4, 0xBE, 0xB9, 0xAF, 0xD0, 0xC3, 0xF1, 0xBE,
	0xB8, 0xB0, 0xA3, 0xDE, 0xDF, 0xDF, 0xDF, 0xF2,
	0xA3, 0x81, 0xC0, 0x80, 0xCD, 0xC7, 0xCF, 0xBD,
	0xB4, 0xB2, 0xA3, 0x86, 0x93, 0xF1, 0x20, 0xFD,
	0x3F, 0x2C, 0xFD, 0x3F, 0x34, 0xFD, 0x3F, 0x3C,
	0xFD, 0x3F, 0xBE, 0xA2, 0x20, 0xFD, 0x31, 0xA0,
	0x2C, 0xFD, 0x31, 0x34, 0xFD, 0x31, 0x3C, 0xFD,
	0x31, 0xBE, 0xBE, 0xBC, 0xBC, 0xB0, 0xA3, 0x8B,
	0x93, 0x2D, 0x55, 0x7D, 0xA2, 0x86, 0x9D, 0x08,
	0xFD, 0x0F, 0xBC, 0xBC, 0xA3, 0x83, 0x92, 0x01,
	0xA9, 0x9E, 0x0E, 0x16, 0x1E, 0xBE, 0xBE, 0x99,
	/* bank 10: 0x0A00 */
	0xA5, 0x2C, 0x54, 0x7C, 0xBA, 0xA5, 0x2C, 0x54,
	0x7C, 0xBD, 0xBD, 0xBC, 0xBC, 0xB1, 0xB6, 0x83,
	0x95, 0xA5, 0xF1, 0x0E, 0x16, 0x1E, 0xB2, 0xA7,
	0x85, 0x95, 0x2A, 0xF0, 0x50, 0x78, 0x87, 0x93,
	0xF1, 0x01, 0xDA, 0xA5, 0xDF, 0xDF, 0xDF, 0xD8,
	0xA4, 0xDF, 0xDF, 0xDF, 0xB0, 0x80, 0xF2, 0xA4,
	0xC3, 0xCB, 0xC5, 0xB2, 0x86, 0x94, 0xBE, 0xBA,
	0xA0, 0xF1, 0x2C, 0xFD, 0x3F, 0x34, 0xFD, 0x3F,
	0x3C, 0xFD, 0x3F, 0xBE, 0xB8, 0xA6, 0x2C, 0xFD,
	0x30, 0x34, 0xFD, 0x30, 0x3C, 0xFD, 0x30, 0xBE,
	0xBA, 0xA4, 0x2C, 0xFD, 0x3B, 0x34, 0xFD, 0x3B,
	0x3C, 0xFD, 0x3B, 0xBC, 0xB2, 0x8E, 0x94, 0xB8,
	0xA6, 0x2D, 0x55, 0x7D, 0xBA, 0xA4, 0x2D, 0x55,
	0x7D, 0xBC, 0xBC, 0xF3, 0xB9, 0xA9, 0x80, 0xCD,
	0xF2, 0xC4, 0xC5, 0xBA, 0xF3, 0xA0, 0xD0, 0xDE,
	0xB1, 0xB4, 0xF7, 0xA7, 0x89, 0x91, 0x72, 0x89,
	0x91, 0x47, 0xB6, 0x97, 0x4A, 0xB9, 0xF2, 0xA9,
	0xD0, 0xFA, 0xF9, 0xD1, 0xD9, 0xF4, 0x0A, 0x97,
	0xD8, 0xF3, 0xBA, 0xA7, 0xF9, 0xDB, 0xFB, 0xD9,
	0xF1, 0xB9, 0xB0, 0x81, 0xA9, 0xC3, 0xF2, 0xC5,
	0xF3, 0xBA, 0xA0, 0xD0, 0xF8, 0xD8, 0xF1, 0xB1,
	0x89, 0xA7, 0xDF, 0xDF, 0xDF, 0xF2, 0xA7, 0xC3,
	0xCB, 0xC5, 0xF1, 0xB2, 0xB5, 0xB9, 0x87, 0x97,
	0xA5, 0x22, 0xF0, 0x48, 0x70, 0x3C, 0x98, 0x40,
	0x68, 0x34, 0x58, 0x99, 0x60, 0xF1, 0xBC, 0xB3,
	0x8E, 0x95, 0xAA, 0x25, 0x4D, 0x75, 0xBC, 0xBC,
	0xB8, 0xB0, 0xB4, 0xA7, 0x88, 0x9F, 0xF7, 0x5A,
	0xF9, 0xD9, 0xF1, 0xBA, 0xA5, 0xDF, 0xDF, 0xDF,
	0xB8, 0xDA, 0xF3, 0xA8, 0xF8, 0x88, 0x9D, 0xD0,
	0x7C, 0xD8, 0xF7, 0xA7, 0x88, 0x9F, 0x52, 0xF9,
	0xD9, 0xF1, 0xBA, 0xA4, 0xDF, 0xDF, 0xDF, 0xB8,
	0xDA, 0xF3, 0xA8, 0x88, 0x9C, 0xD0, 0xDF, 0x68,
	/* bank 11: 0x0B00 */
	0x70, 0x9D, 0x60, 0x70, 0x78, 0xD8, 0xF7, 0xA7,
	0x88, 0x9F, 0x42, 0xF9, 0xBA, 0xA0, 0xD0, 0xF3,
	0xD9, 0xDE, 0xD8, 0xF8, 0xF9, 0xD1, 0xB8, 0xDA,
	0xA8, 0x88, 0x9E, 0xD0, 0x64, 0x68, 0xD8, 0xA8,
	0x84, 0x98, 0xD0, 0xF7, 0x7E, 0xF0, 0xB2, 0xB6,
	0xBA, 0x85, 0x91, 0xA7, 0x24, 0x70, 0x59, 0x44,
	0x69, 0x38, 0x64, 0x48, 0x31, 0x2D, 0x51, 0x79,
	0x87, 0xF1, 0xA1, 0x00, 0x2C, 0x54, 0x7C, 0xF0,
	0x81, 0xA7, 0x04, 0x28, 0x50, 0x78, 0xFD, 0x7F,
	0xF1, 0xA7, 0x87, 0x96, 0x59, 0x91, 0xA1, 0x02,
	0x0E, 0x16, 0x1E, 0xF0, 0x84, 0x91, 0xA7, 0x24,
	0x70, 0x59, 0x44, 0x69, 0x38, 0x64, 0x48, 0x31,
	0x2D, 0x51, 0x79, 0xA2, 0x87, 0x0D, 0x20, 0x59,
	0x70, 0x15, 0x38, 0x40, 0x69, 0xA4, 0xF1, 0x62,
	0xF0, 0x19, 0x31, 0x48, 0xB8, 0xB1, 0xB4, 0xF1,
	0xA6, 0x80, 0xC6, 0xF4, 0xB0, 0x81, 0xF3, 0xA7,
	0xC6, 0xB1, 0x8F, 0x97, 0xF7, 0x02, 0xF9, 0xDA,
	0xF4, 0x0B, 0xC6, 0xD8, 0xB0, 0xF7, 0xA7, 0x88,
	0x9F, 0x52, 0xF9, 0xD9, 0xF4, 0x0B, 0xBF, 0xD8,
	0xF1, 0xB2, 0xB6, 0xA6, 0x82, 0x92, 0x2A, 0xF0,
	0x50, 0xFD, 0x08, 0xF1, 0xA7, 0x84, 0x94, 0x02,
	0xFD, 0x08, 0xB0, 0xB4, 0x86, 0x97, 0x00, 0xB1,
	0xBA, 0xA7, 0x81, 0x61, 0xD9, 0xF4, 0x0B, 0xFC,
	0xD8, 0xF1, 0x41, 0xDA, 0xF4, 0x0B, 0xFC, 0xD8,
	0xF1, 0xB8, 0xB2, 0xA6, 0x82, 0xC0, 0xD8, 0xF1,
	0xB0, 0xB6, 0x86, 0x92, 0xA7, 0x16, 0xFD, 0x04,
	0x0F, 0xFD, 0x04, 0xF0, 0xBA, 0x87, 0x91, 0xA7,
	0x24, 0x58, 0x3D, 0x40, 0x34, 0x49, 0x2D, 0x51,
	0xB2, 0x87, 0xF1, 0xA1, 0x00, 0x2C, 0x54, 0x7C,
	0xF0, 0x81, 0xA7, 0x04, 0x28, 0x50, 0x78, 0xFD,
	0x7F, 0xF1, 0xA7, 0x87, 0x96, 0x59, 0x91, 0xA1,
	0x02, 0x0E, 0x16, 0x1E, 0xD8, 0xF7, 0xB0, 0xB4,
	/* bank 12: 0x0C00 */
	0xBA, 0x88, 0x9E, 0xA7, 0x6A, 0xF9, 0xD9, 0xF4,
	0x0D, 0x1E, 0xD8, 0xF3, 0xB6, 0x90, 0xFC, 0xC2,
	0x00, 0xD9, 0xF4, 0x0D, 0x0E, 0xD8, 0xF0, 0xB1,
	0xB5, 0x8A, 0x9A, 0xA3, 0x2C, 0x50, 0x78, 0xF2,
	0xA5, 0xDE, 0xF8, 0xF8, 0xF1, 0xB5, 0xB2, 0xA7,
	0x83, 0x90, 0x21, 0xDB, 0xB6, 0xB1, 0x80, 0x93,
	0x29, 0xD9, 0xF2, 0xA5, 0xF8, 0xD8, 0xF1, 0xB2,
	0xB6, 0xA7, 0x83, 0x93, 0x31, 0xF3, 0xA5, 0xDE,
	0xD9, 0xF8, 0xF8, 0xD8, 0xF1, 0xA9, 0x89, 0x99,
	0xF0, 0x34, 0x83, 0x38, 0xF1, 0xA7, 0x29, 0xF5,
	0x87, 0x90, 0x18, 0xD9, 0xF3, 0xA5, 0xF8, 0xD8,
	0xA8, 0x80, 0x91, 0xF0, 0x0C, 0x31, 0x14, 0x28,
	0x1C, 0x20, 0x04, 0x39, 0xA7, 0x88, 0x98, 0x04,
	0x28, 0x51, 0x79, 0x1D, 0x30, 0x14, 0x38, 0xB1,
	0xB8, 0x8A, 0xA7, 0xD0, 0x97, 0x2C, 0x50, 0x50,
	0x78, 0x78, 0xBC, 0xBA, 0xB0, 0xA7, 0x85, 0x98,
	0x04, 0x28, 0x50, 0x78, 0xF1, 0x84, 0x97, 0x29,
	0xD9, 0xBE, 0xB8, 0xA5, 0x8D, 0x98, 0x20, 0x2C,
	0x34, 0x3C, 0xBE, 0xBE, 0xBA, 0xAA, 0xDE, 0xDF,
	0xF8, 0xF4, 0x0C, 0xA9, 0xD8, 0xF1, 0xBA, 0xAA,
	0xF8, 0xA7, 0x84, 0x9A, 0x01, 0xD9, 0xAA, 0xDF,
	0xF8, 0xD8, 0xF1, 0xBC, 0xBC, 0x9A, 0xFC, 0xC1,
	0x04, 0xD9, 0xAA, 0xFB, 0xA6, 0xB2, 0x89, 0x96,
	0x42, 0xB4, 0x97, 0xF0, 0x78, 0xA7, 0x86, 0xF1,
	0x38, 0xFD, 0x01, 0xB6, 0x97, 0x01, 0xA6, 0xD0,
	0xB6, 0x9A, 0x52, 0x87, 0xF0, 0x18, 0xD8, 0xB2,
	0xB6, 0xBA, 0xF1, 0xA9, 0x89, 0x93, 0x1A, 0xB0,
	0x87, 0x99, 0x60, 0xB8, 0xA7, 0xD0, 0xFD, 0x42,
	0xBA, 0xF3, 0xA5, 0xF9, 0xF9, 0xB8, 0xF1, 0xDA,
	0xA7, 0xD0, 0xDF, 0xD8, 0xA7, 0xD0, 0xFA, 0xF9,
	0xD1, 0xBA, 0xDA, 0xA7, 0x87, 0x90, 0x6A, 0x66,
	0xB2, 0xA0, 0x87, 0x01, 0x2C, 0xF0, 0xA7, 0x80,
	/* bank 13: 0x0D00 */
	0x90, 0x04, 0x28, 0xFD, 0x7F, 0xF1, 0xA7, 0x87,
	0x96, 0x59, 0x90, 0xA0, 0x02, 0x0E, 0xD8, 0xF0,
	0xBA, 0xB2, 0xB6, 0xA8, 0x80, 0x91, 0x0C, 0x31,
	0x14, 0x28, 0x1C, 0x20, 0x04, 0x39, 0xD8, 0xF7,
	0xB8, 0xB4, 0xB0, 0xA7, 0x9D, 0x88, 0x72, 0xF9,
	0xBC, 0xBD, 0xBE, 0xD9, 0xF4, 0x0E, 0x81, 0xD8,
	0xF2, 0xB8, 0xAD, 0xF8, 0xF9, 0xD1, 0xDA, 0xDE,
	0xBA, 0xA1, 0xDE, 0xAE, 0xDE, 0xF8, 0xD8, 0xF2,
	0xB1, 0xB5, 0xB9, 0xAE, 0xF9, 0xDA, 0xF4, 0x0E,
	0x68, 0xD8, 0xF2, 0x8E, 0xC2, 0xF1, 0xB2, 0x80,
	0x9A, 0xF5, 0xAF, 0x24, 0xD9, 0xF4, 0x0E, 0x68,
	0xD8, 0xF5, 0x44, 0xD9, 0xF4, 0x0E, 0x68, 0xD8,
	0xF5, 0x64, 0xD9, 0xF4, 0x0E, 0x68, 0xD8, 0xF1,
	0xB1, 0xB6, 0x8B, 0x90, 0xAF, 0x2D, 0x55, 0x7D,
	0xB5, 0x8C, 0x9F, 0xAD, 0x0E, 0x16, 0x1E, 0x8B,
	0x9D, 0xAB, 0x2C, 0x54, 0x7C, 0x8D, 0x9F, 0xAA,
	0x2E, 0x56, 0x7E, 0x8A, 0x9C, 0xAA, 0x2C, 0x54,
	0x7C, 0x9B, 0xAC, 0x26, 0x46, 0x66, 0xAF, 0x8D,
	0x9D, 0x00, 0x9C, 0x0D, 0xDB, 0x11, 0x8F, 0x19,
	0xF4, 0x0D, 0x9E, 0xD8, 0x0E, 0x68, 0xD8, 0xF1,
	0xB2, 0x81, 0xB6, 0x90, 0xAF, 0x2D, 0x55, 0x7D,
	0xB1, 0x8F, 0xB5, 0x9F, 0xAF, 0xF5, 0x2C, 0x54,
	0x7C, 0x8F, 0x9F, 0xAD, 0xF1, 0x35, 0xD9, 0xAF,
	0xC2, 0xDA, 0xAF, 0xC4, 0xD8, 0xAD, 0x1D, 0xDA,
	0xAF, 0xC6, 0xD8, 0xB6, 0x8F, 0x9C, 0xAD, 0x1D,
	0xBA, 0xD9, 0xF3, 0xA1, 0xDE, 0xF8, 0xF1, 0xB2,
	0x80, 0xA1, 0xC3, 0xC5, 0xC7, 0xF4, 0x0D, 0xE5,
	0xD8, 0xF3, 0x91, 0xFC, 0xC0, 0x28, 0xDA, 0xA1,
	0xF8, 0xD9, 0xF4, 0x0E, 0x68, 0xD8, 0xF3, 0xB9,
	0xAE, 0xF8, 0xF9, 0xD1, 0xD9, 0xF8, 0xF4, 0x0E,
	0x68, 0xD8, 0xF1, 0xBA, 0xB1, 0xB5, 0xA0, 0x8B,
	0x9E, 0x34, 0xFD, 0x32, 0x54, 0xFD, 0x32, 0x74,
	/* bank 14: 0x0E00 */
	0xFD, 0x32, 0xB2, 0xB6, 0xA3, 0x87, 0xC0, 0xC3,
	0xC5, 0xC7, 0xA2, 0x88, 0xC0, 0xC3, 0xC5, 0xC7,
	0xA4, 0x86, 0xC0, 0xC3, 0xC5, 0xC7, 0xA5, 0x85,
	0xC4, 0xC7, 0xAC, 0x8D, 0xC0, 0xF3, 0xAE, 0xDE,
	0xF8, 0xF4, 0x11, 0x72, 0xD8, 0xF1, 0xA7, 0x83,
	0xC0, 0xC3, 0xC5, 0xC7, 0xA8, 0x82, 0xC0, 0xC3,
	0xC5, 0xC7, 0xA6, 0x84, 0xC0, 0xC3, 0xC5, 0xC7,
	0xA5, 0x85, 0xD0, 0xC0, 0xC3, 0x8D, 0x9D, 0xAF,
	0x39, 0xD9, 0xF4, 0x0E, 0x68, 0xD8, 0xF1, 0x83,
	0xB5, 0x9E, 0xAE, 0x34, 0xFD, 0x0A, 0x54, 0xFD,
	0x0A, 0x74, 0xFD, 0x0A, 0xF2, 0xAF, 0xDE, 0xF8,
	0xF8, 0xF8, 0xB6, 0x81, 0x9F, 0x05, 0xF8, 0xF9,
	0xD1, 0xDA, 0x8F, 0xA1, 0xC0, 0xF4, 0x0E, 0x6F,
	0xD8, 0xF2, 0xBA, 0xAE, 0xF8, 0xF9, 0xD1, 0xDA,
	0xF3, 0xBE, 0xBE, 0xBC, 0xBC, 0xBD, 0xBD, 0xB8,
	0xB0, 0xB4, 0xA5, 0x85, 0x9C, 0x08, 0xBE, 0xBC,
	0xBD, 0xD8, 0xF7, 0xBC, 0xBC, 0xBD, 0xBD, 0xBB,
	0xB4, 0xB0, 0xAF, 0x9E, 0x88, 0x62, 0xF9, 0xBC,
	0xBD, 0xD9, 0xF4, 0x10, 0xF5, 0xD8, 0xF1, 0xBC,
	0xBC, 0xB1, 0x85, 0xBA, 0xA0, 0xC1, 0xC3, 0xC5,
	0xBC, 0xB6, 0xB9, 0xAF, 0x87, 0x90, 0x2D, 0x55,
	0x7D, 0xB5, 0xAD, 0x87, 0x9F, 0x0E, 0x16, 0x1E,
	0xA7, 0x9D, 0x2C, 0x54, 0x7C, 0xAF, 0x8F, 0x2E,
	0x56, 0x7E, 0xAF, 0x98, 0x2C, 0x54, 0x7C, 0xA8,
	0x26, 0x46, 0x66, 0xB6, 0xBA, 0xA0, 0x8E, 0x90,
	0x4C, 0xFD, 0x3B, 0x54, 0xFD, 0x3B, 0x5C, 0xFD,
	0x3B, 0xB9, 0xB3, 0xB6, 0xF1, 0xAF, 0x81, 0x2D,
	0x55, 0x7D, 0xB1, 0xB5, 0xAF, 0x8F, 0x9F, 0xF5,
	0x2C, 0x54, 0x7C, 0xF1, 0xBB, 0xAF, 0x86, 0x98,
	0x49, 0xDB, 0x51, 0x59, 0x9F, 0x69, 0x71, 0x79,
	0xB3, 0xB7, 0x8B, 0x9C, 0x61, 0xDA, 0xAC, 0xDE,
	0xF3, 0xA0, 0xDF, 0xF8, 0xD8, 0xB3, 0xB7, 0xAF,
	/* bank 15: 0x0F00 */
	0x80, 0x90, 0x69, 0xD9, 0xA0, 0xFA, 0xF1, 0xB2,
	0x80, 0xA1, 0xC3, 0xC5, 0xC7, 0xF4, 0x10, 0xDF,
	0xD8, 0xF1, 0xBC, 0xBC, 0xB1, 0xAD, 0x8A, 0x9A,
	0x34, 0xFD, 0x3A, 0x54, 0xFD, 0x3A, 0x74, 0xFD,
	0x3A, 0xBC, 0xB3, 0xB3, 0xAD, 0x8D, 0x9D, 0xF5,
	0x2C, 0x54, 0x7C, 0xF3, 0xA2, 0xDE, 0xF8, 0xF4,
	0x11, 0x00, 0xD8, 0xF1, 0xAA, 0x8D, 0xC1, 0xF2,
	0xA1, 0xF8, 0xF9, 0xD1, 0xDA, 0xF4, 0x0F, 0x66,
	0xD8, 0xF1, 0xAF, 0x8A, 0x9A, 0x21, 0x8F, 0x9F,
	0xF5, 0x00, 0xF1, 0x80, 0x41, 0xD9, 0xF4, 0x0F,
	0x66, 0xD8, 0xF1, 0x91, 0xFC, 0xC0, 0x04, 0xD9,
	0xF4, 0x0F, 0x92, 0xD8, 0xF3, 0xA1, 0xDE, 0xF8,
	0xA0, 0xDF, 0xF8, 0xF4, 0x10, 0xDF, 0xF3, 0x91,
	0xFC, 0xC0, 0x07, 0xD9, 0xF4, 0x0F, 0x92, 0xD8,
	0xF1, 0xA3, 0x87, 0xC0, 0xC3, 0xC5, 0xC7, 0xA4,
	0x88, 0xC0, 0xC3, 0xC5, 0xC7, 0xA5, 0x89, 0xC0,
	0xC3, 0xC5, 0xC7, 0xA6, 0x86, 0xC4, 0xC7, 0xA1,
	0x82, 0xC3, 0xC5, 0xC7, 0xF3, 0xA1, 0xDE, 0xF4,
	0x10, 0xF5, 0xD8, 0xF1, 0xB1, 0xB5, 0xAF, 0x86,
	0x98, 0x29, 0xDB, 0x31, 0x39, 0xF4, 0x10, 0xF5,
	0xD8, 0xF1, 0xBA, 0xB3, 0xB6, 0xA0, 0x81, 0x90,
	0x2D, 0x55, 0x7D, 0xBB, 0xF2, 0xA2, 0xF8, 0xF9,
	0xD1, 0xDA, 0xDE, 0xF4, 0x10, 0xF5, 0xD8, 0xF1,
	0xBA, 0xB3, 0xA3, 0x83, 0xC0, 0xC3, 0xC5, 0xC7,
	0xA2, 0x84, 0xC0, 0xC3, 0xC5, 0xC7, 0xA4, 0x85,
	0xC0, 0xC3, 0xC5, 0xC7, 0xA5, 0x86, 0xC0, 0xC3,
	0xAC, 0x8C, 0xC2, 0xF3, 0xAE, 0xDE, 0xF8, 0xF8,
	0xF4, 0x11, 0x72, 0xD8, 0xF1, 0xB2, 0xBB, 0xA3,
	0x83, 0xC0, 0xC3, 0xC5, 0xC7, 0xA4, 0x82, 0xC0,
	0xC3, 0xC5, 0xC7, 0xA5, 0x84, 0xC0, 0xC3, 0xC5,
	0xC7, 0xA6, 0x85, 0xC0, 0xC3, 0xAC, 0x8C, 0xC4,
	0xB3, 0xB7, 0xAF, 0x85, 0x95, 0x06, 0xFD, 0x0F,
	/* bank 16: 0x1000 */
	0x2E, 0xFD, 0x0F, 0x7E, 0xFD, 0x0F, 0xAF, 0x84,
	0x9F, 0x2A, 0xF0, 0x50, 0x78, 0xFD, 0x0F, 0xF1,
	0x85, 0x95, 0x56, 0xFD, 0x0F, 0x86, 0x96, 0x06,
	0xFD, 0x0F, 0xF0, 0x84, 0x9F, 0xAF, 0x4C, 0x70,
	0xFD, 0x0F, 0xF1, 0x86, 0x96, 0x2E, 0xFD, 0x0F,
	0x84, 0x9F, 0x72, 0xFD, 0x0F, 0xDF, 0xAF, 0x00,
	0x2C, 0x54, 0x7C, 0xAE, 0x8F, 0xC0, 0xAF, 0x8F,
	0x9E, 0x21, 0xD9, 0xAE, 0xC2, 0xD8, 0xAF, 0x41,
	0xD9, 0xAE, 0xC4, 0xD8, 0xAF, 0x61, 0xD9, 0xAE,
	0xC6, 0xD8, 0xAF, 0x8C, 0x9E, 0x61, 0x8B, 0x9C,
	0xDB, 0x61, 0xDA, 0xF4, 0x10, 0xF5, 0xD8, 0xF1,
	0xAB, 0x83, 0x91, 0x28, 0xFD, 0x05, 0x54, 0xFD,
	0x05, 0x7C, 0xFD, 0x05, 0xB8, 0xBD, 0xBD, 0xB5,
	0xA3, 0x8B, 0x95, 0x05, 0x2D, 0x55, 0xBD, 0xB4,
	0xBB, 0xAD, 0x8A, 0x93, 0x4C, 0xFD, 0x3A, 0x54,
	0xFD, 0x3A, 0x5C, 0xFD, 0x3A, 0xB7, 0xAD, 0x8D,
	0x9D, 0xF5, 0x2C, 0x54, 0x7C, 0xF3, 0xA2, 0xDE,
	0xF8, 0xF8, 0xF4, 0x11, 0x00, 0xD8, 0xF2, 0xA1,
	0xF8, 0xF9, 0xD1, 0xD9, 0xF1, 0xAF, 0x8D, 0x9A,
	0x01, 0xF5, 0x8F, 0x90, 0xDB, 0x00, 0xF4, 0x10,
	0xF5, 0xDA, 0xF1, 0xAA, 0x8D, 0xC0, 0xAE, 0x8B,
	0xC1, 0xC3, 0xC5, 0xBE, 0xBE, 0xB9, 0xB0, 0xAA,
	0x83, 0xC3, 0xC5, 0xC7, 0xB3, 0xF3, 0xBA, 0xA2,
	0xD0, 0xDF, 0xF8, 0xBE, 0xBB, 0xF1, 0xA1, 0xDE,
	0xA7, 0x83, 0xC0, 0xC3, 0xC5, 0xC7, 0xA8, 0x84,
	0xC0, 0xC3, 0xC5, 0xC7, 0xA9, 0x85, 0xC0, 0xC3,
	0xC5, 0xC7, 0xA6, 0x86, 0xD0, 0xC0, 0xC3, 0xA2,
	0x81, 0xC3, 0xC5, 0xC7, 0xF4, 0x10, 0xF5, 0xF1,
	0xBB, 0xB3, 0xA3, 0xDE, 0xDF, 0xDF, 0xDF, 0xA4,
	0x8C, 0xC4, 0xC5, 0xC5, 0xC5, 0xA5, 0xDE, 0xDF,
	0xDF, 0xDF, 0xA6, 0xDE, 0xDF, 0xD8, 0xF3, 0xB9,
	0xAE, 0xDF, 0xBA, 0xAE, 0xDE, 0xBB, 0xA2, 0xDE,
	/* bank 17: 0x1100 */
	0xD8, 0xF3, 0xA2, 0xF8, 0xF9, 0xD1, 0xD9, 0xF4,
	0x11, 0x70, 0xD8, 0xF1, 0xAF, 0x8D, 0x9D, 0x49,
	0xDA, 0xC3, 0xC5, 0xD9, 0xC5, 0xC3, 0xD8, 0xAF,
	0x9F, 0x69, 0xD0, 0xDA, 0xC7, 0xD9, 0x8F, 0xC3,
	0x8D, 0xAF, 0xC7, 0xD8, 0xB9, 0xA9, 0x8F, 0x9F,
	0xF0, 0x54, 0x78, 0xF1, 0xFD, 0x0F, 0xA6, 0xB1,
	0x89, 0xC2, 0xB3, 0xAF, 0x8F, 0x9F, 0x2E, 0xFD,
	0x11, 0xB1, 0xB5, 0xA9, 0x89, 0x9F, 0x2C, 0xF3,
	0xAE, 0xDF, 0xF8, 0xF8, 0xF4, 0x13, 0x2E, 0xD8,
	0xF1, 0xAD, 0x86, 0x99, 0x06, 0xFD, 0x10, 0xDF,
	0xF8, 0xFD, 0x0F, 0xAD, 0x8D, 0x9D, 0x4C, 0xBB,
	0xB3, 0xAD, 0x8F, 0x9D, 0x2A, 0xFD, 0x0F, 0xB7,
	0x92, 0xFC, 0xC0, 0x04, 0xD9, 0xF4, 0x0F, 0x32,
	0xD8, 0xFC, 0xC0, 0x08, 0xD9, 0xF4, 0x10, 0x8D,
	0xD8, 0xF1, 0xD8, 0xF3, 0xBA, 0xB2, 0xB6, 0xAE,
	0xF8, 0xF9, 0xD1, 0xD9, 0xF4, 0x13, 0x2C, 0xD8,
	0xF1, 0xAF, 0xDE, 0xF9, 0xFD, 0x0F, 0x80, 0x90,
	0x2C, 0x54, 0x7C, 0xA0, 0x2A, 0xF0, 0x50, 0x78,
	0xFD, 0x0F, 0xF1, 0xA2, 0x82, 0x9C, 0x00, 0x24,
	0x44, 0x64, 0xA9, 0x8F, 0x94, 0xF0, 0x04, 0xFD,
	0x0F, 0x0C, 0x30, 0xFD, 0x0F, 0x1C, 0x95, 0x20,
	0x48, 0xFD, 0x0F, 0xF1, 0x99, 0xC1, 0x2C, 0x54,
	0x7C, 0xAA, 0x82, 0x99, 0x02, 0xFD, 0x0F, 0x2E,
	0xFD, 0x0F, 0x56, 0xFD, 0x0F, 0x7E, 0xFD, 0x0F,
	0xAC, 0x83, 0x9F, 0xF0, 0x04, 0x28, 0x50, 0x78,
	0xFD, 0x0F, 0x8C, 0x90, 0xF1, 0x21, 0xF5, 0x8C,
	0x9C, 0x2C, 0xF1, 0xAB, 0xDE, 0xF8, 0xFD, 0x0F,
	0x8B, 0xC1, 0xAF, 0xDE, 0xF1, 0x89, 0xAF, 0x9F,
	0xFC, 0xC0, 0x00, 0xD9, 0xC1, 0x8A, 0xC1, 0x82,
	0xC1, 0xD8, 0xFC, 0xC0, 0x04, 0xD9, 0xC3, 0x8A,
	0xC3, 0x82, 0xC3, 0xD8, 0xFC, 0xC0, 0x08, 0xD9,
	0xC5, 0x8A, 0xC5, 0x82, 0xC5, 0xD8, 0xFC, 0xC0,
	/* bank 18: 0x1200 */
	0x0C, 0xD9, 0xC7, 0x8A, 0xC7, 0x82, 0xC7, 0xD8,
	0xFC, 0xC0, 0x10, 0xD9, 0xF4, 0x12, 0xE6, 0xD8,
	0xF1, 0x8B, 0xAB, 0xD0, 0xC0, 0x9F, 0x2E, 0xFD,
	0x0F, 0xA0, 0xDE, 0xAB, 0xD0, 0x90, 0x65, 0xA0,
	0x8F, 0x9F, 0x4A, 0xFD, 0x0F, 0xAB, 0x8B, 0x90,
	0x00, 0xB9, 0xA9, 0xC1, 0xF3, 0xAE, 0xDF, 0xF8,
	0xF4, 0x13, 0x2E, 0xD8, 0xF1, 0xBA, 0xB1, 0xB6,
	0x89, 0xAB, 0xC1, 0xB2, 0xAF, 0xD0, 0x8B, 0x9F,
	0x3E, 0xFD, 0x0F, 0x5A, 0xFD, 0x0F, 0x9F, 0xFC,
	0xC0, 0x00, 0xD9, 0xF1, 0x8F, 0xA2, 0xC6, 0xD8,
	0xFC, 0xC0, 0x04, 0xD9, 0x8F, 0xA2, 0xC7, 0x84,
	0xAB, 0xD0, 0xC0, 0xAF, 0x8A, 0x9B, 0x1E, 0xFD,
	0x0F, 0x36, 0xFD, 0x0F, 0xA4, 0x8F, 0x30, 0xAA,
	0x9A, 0x40, 0xD8, 0x9F, 0xFC, 0xC0, 0x08, 0xD9,
	0x8F, 0xA2, 0xD0, 0xC6, 0x84, 0xAB, 0xD0, 0xC2,
	0xAF, 0x8A, 0x9B, 0x1E, 0xFD, 0x0F, 0x56, 0xFD,
	0x0F, 0xA4, 0x8F, 0x34, 0xAA, 0x9A, 0x40, 0x84,
	0xAB, 0xD0, 0xC4, 0xAF, 0x8A, 0x9B, 0x3E, 0xFD,
	0x0F, 0x56, 0xFD, 0x0F, 0xA4, 0xD0, 0x8F, 0x30,
	0xAA, 0x9A, 0x4C, 0xD8, 0x9F, 0xFC, 0xC0, 0x0C,
	0xD9, 0x8F, 0xA2, 0xD0, 0xC7, 0x84, 0xAB, 0xD0,
	0xC6, 0xAF, 0x8A, 0x9B, 0x1E, 0xFD, 0x0F, 0x76,
	0xFD, 0x0F, 0xA4, 0xD0, 0x8F, 0x34, 0xAA, 0x9A,
	0x40, 0x85, 0xAB, 0xD0, 0xC0, 0xAF, 0x8A, 0x9B,
	0x3E, 0xFD, 0x0F, 0x76, 0xFD, 0x0F, 0xA5, 0x8F,
	0x30, 0xAA, 0x9A, 0x4C, 0x85, 0xAB, 0xD0, 0xC2,
	0xAF, 0x8A, 0x9B, 0x5E, 0xFD, 0x0F, 0x76, 0xFD,
	0x0F, 0xA5, 0x8F, 0x34, 0xAA, 0xD0, 0x9A, 0x50,
	0xD8, 0xAF, 0xF8, 0xF4, 0x11, 0xDC, 0xF1, 0xD8,
	0x8B, 0x9C, 0xAF, 0x2A, 0xFD, 0x0F, 0x8A, 0x9F,
	0xB9, 0xAF, 0x02, 0xFD, 0x0F, 0x26, 0xFD, 0x0F,
	0x46, 0xFD, 0x0F, 0x66, 0xFD, 0x0F, 0x83, 0xB5,
	/* bank 19: 0x1300 */
	0x9F, 0xBA, 0xA3, 0x00, 0x2C, 0x54, 0x7C, 0xB6,
	0x82, 0x92, 0xA0, 0x31, 0xD9, 0xAD, 0xC3, 0xDA,
	0xAD, 0xC5, 0xD8, 0x8D, 0xA0, 0x39, 0xDA, 0x82,
	0xAD, 0xC7, 0xD8, 0xF3, 0x9E, 0xFC, 0xC0, 0x04,
	0xD9, 0xF4, 0x0E, 0x24, 0xD8, 0xFC, 0xC0, 0x08,
	0xD9, 0xF4, 0x0F, 0xDB, 0xD8, 0xF1, 0xD8, 0xF1,
	0xB9, 0xB1, 0xB5, 0xA9, 0xDE, 0xF8, 0x89, 0x99,
	0xAF, 0x31, 0xD9, 0xF4, 0x13, 0x82, 0xD8, 0xF1,
	0x85, 0xAF, 0x29, 0xD9, 0x84, 0xA9, 0xC2, 0xD8,
	0x85, 0xAF, 0x49, 0xD9, 0x84, 0xA9, 0xC4, 0xD8,
	0x85, 0xAF, 0x69, 0xD9, 0x84, 0xA9, 0xC6, 0xD8,
	0x89, 0xAF, 0x39, 0xDA, 0x8E, 0xA9, 0x50, 0xF4,
	0x13, 0x82, 0xD8, 0xF1, 0x89, 0xAA, 0x7C, 0xFD,
	0x02, 0x9A, 0x68, 0xD8, 0xF1, 0xAA, 0xFB, 0xDA,
	0x89, 0x99, 0xAF, 0x26, 0xFD, 0x0F, 0x8F, 0x95,
	0x25, 0x89, 0x9F, 0xA9, 0x12, 0xFD, 0x0F, 0xF4,
	0x13, 0x6B, 0xD8, 0xF3, 0x9E, 0xFC, 0xC1, 0x04,
	0xD9, 0xF4, 0x12, 0x33, 0xD8, 0xFC, 0xC1, 0x08,
	0xD9, 0xF4, 0x11, 0x47, 0xD8, 0xF1, 0xBE, 0xBC,
	0xBC, 0xBD, 0xBD, 0xF7, 0xB8, 0xB4, 0xB0, 0xAC,
	0x84, 0x9D, 0x12, 0xF9, 0xF2, 0xBC, 0xBC, 0x8A,
	0xA7, 0xD0, 0xD9, 0xC3, 0xDA, 0xC5, 0xD8, 0xBC,
	0xBD, 0xBD, 0xF3, 0x8F, 0xA8, 0xC0, 0xF9, 0xAC,
	0x84, 0x97, 0xF5, 0x1A, 0xF1, 0xF8, 0xF9, 0xD1,
	0xDA, 0xA8, 0xDE, 0xD8, 0x95, 0xFC, 0xC1, 0x03,
	0xD9, 0xA8, 0xDE, 0xD8, 0xBC, 0xBC, 0xF1, 0x98,
	0xFC, 0xC0, 0x1C, 0xDB, 0x95, 0xFC, 0xC0, 0x03,
	0xA5, 0xDE, 0xA4, 0xDE, 0xD8, 0xAC, 0x88, 0x95,
	0x00, 0xD1, 0xD9, 0xA5, 0xF8, 0xD8, 0xA4, 0xFC,
	0x80, 0x04, 0x88, 0x95, 0xA4, 0xFC, 0x08, 0x04,
	0x20, 0xF7, 0xBC, 0xBD, 0xB5, 0xAC, 0x84, 0x9F,
	0xF6, 0x02, 0xF8, 0xF9, 0xD1, 0xDB, 0x84, 0x93,
	/* bank 20: 0x1400 */
	0xF7, 0x6A, 0xF9, 0xD9, 0xF3, 0xBC, 0xBC, 0xA8,
	0x88, 0x92, 0x18, 0xBC, 0xD8, 0xBC, 0xBC, 0xB4,
	0xA8, 0x88, 0x9E, 0x08, 0xF4, 0xBE, 0xA1, 0xD0,
	0xC1, 0xBC, 0xF7, 0xBE, 0xBE, 0xB5, 0xAC, 0x84,
	0x93, 0x6A, 0xF9, 0xBD, 0xBD, 0xB4, 0xD9, 0xF2,
	0xAC, 0x8C, 0x97, 0x18, 0xF6, 0x84, 0x9C, 0x02,
	0xF8, 0xF9, 0xDB, 0xD1, 0xF1, 0xA3, 0xDE, 0xA5,
	0xDF, 0xD8, 0xF7, 0xBE, 0xBD, 0xA7, 0x9D, 0x88,
	0x7A, 0xF9, 0xD9, 0xF4, 0x15, 0xA7, 0xD8, 0xF1,
	0xBE, 0xBE, 0xBD, 0xBD, 0xBC, 0xBC, 0x93, 0xA3,
	0xFC, 0xC0, 0x08, 0xD9, 0xDE, 0xD8, 0x95, 0xFC,
	0xC0, 0x08, 0xD9, 0xA5, 0xDF, 0xF8, 0xD8, 0x95,
	0xFC, 0xC1, 0x04, 0xDB, 0xFC, 0xC0, 0x04, 0xA3,
	0xDE, 0xD8, 0x93, 0xFC, 0xC0, 0x04, 0xDA, 0xF4,
	0x15, 0x52, 0xD8, 0xF1, 0xAC, 0x82, 0x96, 0x2D,
	0x55, 0x7D, 0x8C, 0x9C, 0x34, 0x18, 0xF5, 0xAC,
	0x95, 0x1C, 0xD9, 0xF1, 0xA3, 0xDE, 0xA5, 0xDF,
	0xF4, 0x15, 0x52, 0xD8, 0xF7, 0xBC, 0xBD, 0xAC,
	0x88, 0x9F, 0x56, 0xF9, 0xBC, 0xBC, 0xBD, 0xBD,
	0xD9, 0xF4, 0x14, 0xB7, 0xD8, 0xF1, 0xAC, 0x83,
	0x90, 0x45, 0xF9, 0xD9, 0xA0, 0xF8, 0xAC, 0x8C,
	0x9C, 0x06, 0xD2, 0xA1, 0x91, 0x00, 0x2C, 0x87,
	0xD4, 0xF0, 0xA7, 0x8C, 0x9C, 0x20, 0xD3, 0xD8,
	0xF3, 0xBD, 0x98, 0xFC, 0xC0, 0x04, 0xBD, 0xBD,
	0xD9, 0xF1, 0xAC, 0x83, 0x98, 0x49, 0xDB, 0xF9,
	0xD9, 0xA8, 0xFA, 0xA4, 0x84, 0x90, 0x2C, 0x54,
	0x7C, 0xD8, 0xA0, 0x83, 0x93, 0xFC, 0x08, 0x04,
	0x20, 0xFC, 0xC0, 0x08, 0xDA, 0xF4, 0x15, 0x52,
	0xD8, 0xF1, 0xAC, 0x81, 0x91, 0x02, 0xFD, 0x17,
	0x87, 0xD4, 0xFD, 0x47, 0xF0, 0x81, 0x09, 0x09,
	0x09, 0x9C, 0x00, 0x00, 0xF1, 0x93, 0xFC, 0x4C,
	0x08, 0x00, 0xD8, 0xA9, 0x82, 0x96, 0x01, 0x93,
	/* bank 21: 0x1500 */
	0xFC, 0xC0, 0x08, 0xDB, 0x95, 0xFC, 0xC1, 0x00,
	0xA3, 0xDE, 0xAC, 0x89, 0x93, 0xF5, 0x18, 0xF1,
	0xA5, 0xDF, 0xF8, 0xD8, 0xF1, 0x83, 0xAC, 0xC4,
	0xF8, 0x8C, 0x93, 0xA8, 0xFC, 0x30, 0x08, 0x10,
	0xFC, 0xC0, 0x08, 0xDB, 0x95, 0xFC, 0xC1, 0x04,
	0xD9, 0xA3, 0xDE, 0xA5, 0xDF, 0xA4, 0x84, 0x34,
	0xFD, 0x08, 0x54, 0xFD, 0x08, 0x74, 0xFD, 0x08,
	0xA9, 0x94, 0xF5, 0x2C, 0x54, 0x7C, 0xF1, 0xDB,
	0xAC, 0x87, 0x99, 0x49, 0x51, 0x59, 0x84, 0xAB,
	0xC3, 0xC5, 0xC7, 0x82, 0xA6, 0xC0, 0xF3, 0xAA,
	0xDF, 0xF8, 0xF1, 0xD8, 0x93, 0xFC, 0xC0, 0x00,
	0xD9, 0x86, 0xA2, 0xC3, 0xC5, 0xC7, 0xA3, 0xDE,
	0xF8, 0xA0, 0xDE, 0xA1, 0xDE, 0xDF, 0xA7, 0xDE,
	0xDF, 0xA8, 0xDF, 0xA4, 0xDF, 0xDF, 0xDF, 0xD8,
	0xF1, 0x9A, 0xFC, 0xC1, 0x04, 0xD9, 0xAC, 0x82,
	0x96, 0x01, 0xF3, 0x81, 0xAA, 0xC4, 0xDB, 0xF5,
	0xAC, 0x8C, 0x9A, 0x18, 0xF3, 0xAA, 0xF9, 0xD8,
	0xAC, 0x8A, 0x9A, 0x41, 0xD1, 0xAA, 0xD0, 0xC0,
	0xD9, 0xF2, 0xAC, 0x85, 0x9A, 0x41, 0xDB, 0xD1,
	0xBC, 0xBD, 0xBE, 0xF4, 0x15, 0xA7, 0xD8, 0xF3,
	0xBC, 0xBD, 0xBE, 0xA5, 0x85, 0x9C, 0x10, 0xD8,
	0xF1, 0xBA, 0xB2, 0xB6, 0xF3, 0x92, 0xFC, 0xC3,
	0x07, 0xD9, 0xF2, 0xA5, 0xDE, 0xD8, 0xF2, 0xA7,
	0x82, 0x95, 0x65, 0xD1, 0x85, 0xA2, 0xD0, 0xC1,
	0xD9, 0xB5, 0xA7, 0x86, 0x93, 0x31, 0xDB, 0xD1,
	0xF4, 0x15, 0xD4, 0xD8, 0xF3, 0xB8, 0xB0, 0xB4,
	0xA5, 0x85, 0x9C, 0x18, 0xD8, 0xF1, 0xBA, 0xB2,
	0xB6, 0x81, 0x96, 0xA1, 0xF8, 0xF9, 0xB9, 0xA6,
	0xDA, 0xC3, 0xC5, 0xC7, 0xD9, 0x2D, 0x4D, 0x6D,
	0xD8, 0xBA, 0x88, 0xA8, 0xF8, 0xF9, 0xA7, 0xDA,
	0xC3, 0xC5, 0xC7, 0xD9, 0x2D, 0x4D, 0x6D, 0xD8,
	0xF2, 0xB0, 0xB9, 0xA3, 0xFA, 0xF9, 0xD1, 0xDA,
	/* bank 22: 0x1600 */
	0xB8, 0x8F, 0xA7, 0xC0, 0xF9, 0xB5, 0x87, 0x93,
	0xF6, 0x0A, 0xF2, 0xB4, 0xA4, 0x84, 0x97, 0x24,
	0xA4, 0x84, 0x9E, 0x3C, 0xD8, 0xF7, 0xB9, 0xB0,
	0xB5, 0xA6, 0x88, 0x95, 0x5A, 0xF9, 0xDA, 0xF1,
	0xAB, 0xF8, 0xD8, 0xB8, 0xB4, 0xF3, 0x98, 0xFC,
	0xC0, 0x04, 0xDA, 0xF4, 0x16, 0x7E, 0xD8, 0xF2,
	0xA9, 0xD0, 0xF8, 0x89, 0x9B, 0xA7, 0x51, 0xD9,
	0xA9, 0xD0, 0xDE, 0xA4, 0x84, 0x9E, 0x2C, 0xD8,
	0xA8, 0xFA, 0x88, 0x9A, 0xA7, 0x29, 0xD9, 0xA8,
	0xDF, 0xA4, 0x84, 0x9D, 0x34, 0xD8, 0xA8, 0xD0,
	0xF8, 0x88, 0x9A, 0xA7, 0x51, 0xD9, 0xA8, 0xD0,
	0xDE, 0xA4, 0x84, 0x9D, 0x2C, 0xD8, 0xA8, 0xD0,
	0xFA, 0x88, 0x9A, 0xA7, 0x79, 0xD9, 0xA8, 0xD0,
	0xDF, 0xA4, 0x84, 0x9D, 0x24, 0xD8, 0xF3, 0xA9,
	0xD0, 0xF8, 0x89, 0x9B, 0xA7, 0x51, 0xD9, 0xA9,
	0xD0, 0xDE, 0xA4, 0x84, 0x9C, 0x2C, 0xD8, 0xF7,
	0xA7, 0x88, 0x9F, 0x52, 0xF9, 0xD9, 0xF4, 0x16,
	0xBB, 0xD8, 0xF1, 0xB9, 0xA2, 0xFA, 0xF3, 0xB8,
	0xA9, 0xD0, 0xFA, 0x89, 0x9B, 0xA7, 0x79, 0xD9,
	0xA9, 0xD0, 0xDF, 0xA4, 0x84, 0x9C, 0x24, 0xD8,
	0xA7, 0x84, 0x9D, 0xF7, 0x12, 0xF9, 0xD9, 0xF4,
	0x16, 0xBB, 0xD8, 0xF2, 0xA8, 0xD0, 0xF8, 0x88,
	0x9A, 0xA7, 0x51, 0xD9, 0xA8, 0xD0, 0xDE, 0xA4,
	0x84, 0x9D, 0x2C, 0xD8, 0xF7, 0xA7, 0x88, 0x9F,
	0x42, 0xF9, 0xD9, 0xF4, 0x17, 0x02, 0xD8, 0xF3,
	0xA9, 0xF8, 0x89, 0x9B, 0xA7, 0x01, 0xD9, 0xA9,
	0xDE, 0xA4, 0x84, 0x9C, 0x3C, 0xD8, 0xA9, 0xFA,
	0x89, 0x9B, 0xA7, 0x29, 0xD9, 0xA9, 0xDF, 0xA4,
	0x84, 0x9C, 0x34, 0xD8, 0xF2, 0xA9, 0xFA, 0x89,
	0x9B, 0xA7, 0x29, 0xD9, 0xA9, 0xDF, 0xA4, 0x84,
	0x9E, 0x34, 0xD8, 0xA9, 0xD0, 0xFA, 0x89, 0x9B,
	0xA7, 0x79, 0xD9, 0xA9, 0xD0, 0xDF, 0xA4, 0x84,
	/* bank 23: 0x1700 */
	0x9E, 0x24, 0xD8, 0xF1, 0xA7, 0xDE, 0xF2, 0x84,
	0xCA, 0x97, 0xA4, 0x24, 0xA5, 0x94, 0xF6, 0x0A,
	0xF7, 0x85, 0x02, 0xF8, 0xF9, 0xD1, 0xD9, 0xF6,
	0x9B, 0x02, 0xD8, 0xA7, 0xB1, 0x82, 0x95, 0x62,
	0xF8, 0xF9, 0xD1, 0xD9, 0xF4, 0x18, 0x5F, 0xD8,
	0xF0, 0xB0, 0x85, 0xA4, 0xD0, 0xC0, 0xDD, 0xF2,
	0xC0, 0xDC, 0xF6, 0xA7, 0x9F, 0x02, 0xF9, 0xD9,
	0xF3, 0xA5, 0xDE, 0xDA, 0xF0, 0xDD, 0xF2, 0xC8,
	0xDC, 0xD8, 0x85, 0x95, 0xA5, 0x00, 0xD9, 0x86,
	0xF0, 0xDD, 0xF1, 0xC2, 0xC4, 0xC6, 0xDC, 0xD8,
	0xF2, 0x85, 0x00, 0xD9, 0x80, 0xF0, 0xDD, 0xF2,
	0xCC, 0xC6, 0xCE, 0xDC, 0xD8, 0x85, 0x00, 0xD9,
	0xB1, 0x89, 0xF0, 0xDD, 0xF2, 0xC2, 0xCA, 0xC4,
	0xDC, 0xD8, 0xB0, 0x85, 0x00, 0xD9, 0x81, 0xF0,
	0xDD, 0xF2, 0xC6, 0xCE, 0x82, 0xC0, 0xC8, 0xDC,
	0xD8, 0x85, 0x00, 0xB1, 0xD9, 0x86, 0xF0, 0xDD,
	0xF1, 0xC2, 0xC4, 0xC6, 0xDC, 0xD8, 0xB0, 0xF2,
	0x85, 0x00, 0xD9, 0xB2, 0x87, 0xF0, 0xDD, 0xF1,
	0xC2, 0xC4, 0xC6, 0xF2, 0xB2, 0x86, 0xC4, 0xDC,
	0xD8, 0xB0, 0x85, 0x00, 0xB1, 0xD9, 0x8F, 0xF0,
	0xDD, 0xF2, 0xC2, 0xC4, 0xC6, 0xDC, 0xD8, 0xB0,
	0x85, 0x00, 0x00, 0xD9, 0x82, 0xF0, 0xDD, 0xF2,
	0xC2, 0xCA, 0xC4, 0xDC, 0xD8, 0x85, 0x00, 0xD9,
	0x85, 0xF0, 0xDD, 0xF1, 0xC2, 0xC4, 0xC6, 0xDC,
	0xD8, 0xF2, 0x85, 0x00, 0xD9, 0xB1, 0x8A, 0xF0,
	0xDD, 0xF1, 0xC2, 0xC4, 0xC6, 0xDC, 0xD8, 0xB0,
	0xF2, 0x85, 0x00, 0xD9, 0xB1, 0xF0, 0xDD, 0xF1,
	0x82, 0xC4, 0xDC, 0xD8, 0xB0, 0xF3, 0xA5, 0xF8,
	0xF9, 0xD1, 0xD9, 0xF4, 0x18, 0x26, 0xD8, 0xF3,
	0x85, 0x95, 0xA5, 0x00, 0x00, 0xD9, 0xBE, 0xF2,
	0xBA, 0xAE, 0xDE, 0xBE, 0xBE, 0xBC, 0xB2, 0x81,
	0xF0, 0xDD, 0xF3, 0xC8, 0xDC, 0xBC, 0xBC, 0xD8,
	/* bank 24: 0x1800 */
	0xB0, 0xB8, 0x85, 0xA5, 0x00, 0xD9, 0xF2, 0xBE,
	0xBE, 0xAA, 0xDE, 0xBE, 0xBC, 0xBC, 0x8A, 0xF0,
	0xDD, 0xF3, 0xC0, 0xDC, 0xBC, 0xD8, 0x85, 0xA5,
	0x00, 0xD9, 0xB9, 0xF2, 0xA3, 0xD0, 0xDE, 0xB2,
	0x85, 0xF0, 0xDD, 0xF3, 0xC8, 0xDC, 0xD8, 0xF2,
	0xB1, 0x86, 0xB9, 0xAF, 0xC3, 0xC5, 0xC7, 0xF2,
	0xB9, 0xA3, 0xDF, 0xB8, 0xB0, 0xB4, 0xA7, 0x84,
	0x9D, 0xF7, 0x1A, 0xF9, 0xD9, 0xF4, 0x18, 0x4D,
	0xD8, 0xF1, 0xB9, 0xB1, 0xB5, 0xA6, 0x83, 0x9B,
	0x61, 0xD9, 0xF4, 0x18, 0x5F, 0xD8, 0xF6, 0xB8,
	0xB0, 0xB4, 0xA7, 0x84, 0x94, 0x5A, 0xF8, 0xF9,
	0xD1, 0xDA, 0xFE, 0xF1, 0xB9, 0xAB, 0xDE, 0xD8,
	0xF1, 0xB8, 0xB0, 0xB4, 0xA7, 0x88, 0x9C, 0xF7,
	0x6A, 0xF9, 0xD9, 0xFF, 0xD8, 0xF1, 0xBB, 0xAA,
	0xF9, 0xDA, 0xFF, 0xD8, 0xB3, 0x8A, 0xC2, 0xB9,
	0xA2, 0xF8, 0xB2, 0x84, 0xBB, 0xAA, 0xD0, 0xC1,
	0x8B, 0xB7, 0x9A, 0xBA, 0xA7, 0x3A, 0x93, 0xF0,
	0x31, 0x31, 0x20, 0xD3, 0x8F, 0xB6, 0x9C, 0xAB,
	0x01, 0x29, 0x51, 0x79, 0xAF, 0xC2, 0xC5, 0xC7,
	0x8B, 0x9B, 0xF1, 0x04, 0xFD, 0x01, 0x87, 0xD4,
	0x8E, 0x9D, 0xAB, 0xF0, 0x01, 0x29, 0x51, 0x79,
	0xAE, 0xC2, 0xC5, 0xC7, 0x8B, 0x9B, 0xF1, 0x04,
	0xFD, 0x01, 0xB3, 0x81, 0xB7, 0x9A, 0xBB, 0xA7,
	0x7A, 0x93, 0xF0, 0x71, 0x71, 0x60, 0x85, 0x94,
	0x01, 0x29, 0x51, 0x79, 0xA5, 0xC2, 0xC5, 0xC7,
	0x87, 0x97, 0xF1, 0x04, 0xFD, 0x01, 0x81, 0x9A,
	0xA7, 0x1A, 0x93, 0xF0, 0x11, 0x11, 0x00, 0x8E,
	0x9B, 0x01, 0x29, 0x51, 0x79, 0xAE, 0xC2, 0xC5,
	0xC7, 0x87, 0x97, 0xF1, 0x04, 0xFD, 0x01, 0x83,
	0xA3, 0xC2, 0xC5, 0xC7, 0x8A, 0xC7, 0xF2, 0xB2,
	0x8B, 0xB6, 0x9B, 0xBA, 0xAB, 0x71, 0xD9, 0xD0,
	0xF8, 0xF5, 0x8F, 0x9F, 0xB9, 0xA6, 0x78, 0x8E,
	/* bank 25: 0x1900 */
	0x9E, 0x7C, 0xF1, 0xB1, 0x8E, 0xB5, 0x96, 0x45,
	0x6D, 0x8B, 0xA6, 0xD0, 0x32, 0x3E, 0x8E, 0xAE,
	0xD0, 0x50, 0x7C, 0xF4, 0x19, 0x8A, 0xD8, 0xF1,
	0xB1, 0x8E, 0xBB, 0xA1, 0xC5, 0xC7, 0xB3, 0xB7,
	0x81, 0x91, 0xAF, 0xD0, 0x50, 0xFD, 0x02, 0xB6,
	0x9B, 0xFC, 0xC2, 0x00, 0x8F, 0xB5, 0x9B, 0xD9,
	0x56, 0xDA, 0x5E, 0xD8, 0xF3, 0xBA, 0xAB, 0xD0,
	0xDE, 0xF8, 0xF1, 0xB7, 0x91, 0xA7, 0x69, 0xDB,
	0xB1, 0x8E, 0x09, 0xF3, 0xAB, 0xD0, 0xDE, 0xF1,
	0xB3, 0x86, 0xB9, 0xA4, 0xC0, 0xD8, 0xB8, 0xB0,
	0xB4, 0xA7, 0x88, 0x9D, 0xF7, 0x62, 0xF9, 0xD9,
	0xF1, 0xB9, 0xA2, 0xDE, 0xF4, 0x19, 0x84, 0xD8,
	0xF1, 0xB1, 0x8A, 0xB5, 0x92, 0x01, 0xD9, 0xF4,
	0x19, 0x84, 0xD8, 0xF1, 0x81, 0xB7, 0x91, 0x09,
	0xDB, 0x31, 0xF4, 0x19, 0x84, 0xD8, 0xF3, 0xA4,
	0xD0, 0xB0, 0x8D, 0xB4, 0x94, 0x1D, 0xF1, 0xB9,
	0xA2, 0xDE, 0xF4, 0x42, 0xD8, 0xF2, 0xBA, 0xAB,
	0xD0, 0xDE, 0xD8, 0xF3, 0xBB, 0xB3, 0xB7, 0x90,
	0xA2, 0x82, 0x00, 0xF2, 0x10, 0xF1, 0x80, 0x9D,
	0xAD, 0xD0, 0x7C, 0xF2, 0xA2, 0xFA, 0xF9, 0xD1,
	0xF1, 0xB9, 0xAC, 0xD9, 0xDE, 0xDA, 0xF8, 0xD8,
	0xF2, 0xBB, 0xAF, 0x92, 0x50, 0x8F, 0x0D, 0xDB,
	0xF1, 0xB1, 0x8C, 0xB5, 0x9C, 0x21, 0xD9, 0xF5,
	0xB3, 0x85, 0xB7, 0x95, 0xB9, 0xA6, 0x78, 0x8E,
	0x9E, 0x7C, 0xF1, 0xB1, 0x8D, 0xB5, 0x9D, 0xAD,
	0x1A, 0xF0, 0x96, 0x40, 0x9D, 0x3C, 0x96, 0x48,
	0xD8, 0xF1, 0xB1, 0x8E, 0xB5, 0x9D, 0xB9, 0xA6,
	0x2A, 0x8D, 0x96, 0x05, 0xD9, 0xF4, 0x1A, 0x09,
	0xD8, 0xF2, 0xB3, 0x80, 0xB7, 0x92, 0xBB, 0xAF,
	0x50, 0xF8, 0x8F, 0x0D, 0xDB, 0xF1, 0xB1, 0x8C,
	0xB5, 0x9C, 0xB9, 0xA6, 0x21, 0xF4, 0x1A, 0x09,
	0xD8, 0xF1, 0xB3, 0x8E, 0xBB, 0xA8, 0xD0, 0xC4,
	/* bank 26: 0x1A00 */
	0xC7, 0xF3, 0xB9, 0xAC, 0xD0, 0xDE, 0xF4, 0x1A,
	0x1A, 0xD8, 0xF1, 0xB3, 0x85, 0xBB, 0xA8, 0xD0,
	0xC4, 0xC7, 0xF3, 0xB9, 0xAC, 0xD0, 0xDE, 0xF8,
	0xDF, 0xF8, 0xD8, 0xF3, 0xB5, 0x9C, 0xFC, 0xC3,
	0x04, 0xDB, 0xFC, 0xC2, 0x00, 0xD9, 0xF2, 0xAC,
	0xD0, 0xDE, 0xD8, 0xF2, 0xBB, 0xAF, 0xB7, 0x92,
	0xB3, 0x82, 0x19, 0x80, 0xA2, 0xD9, 0x26, 0xF3,
	0xA7, 0xD0, 0xDF, 0xD8, 0xF1, 0xAF, 0x89, 0x98,
	0x19, 0xA9, 0x80, 0xD9, 0x38, 0xD8, 0xAF, 0x89,
	0x39, 0xA9, 0x80, 0xDA, 0x3C, 0xD8, 0xAF, 0x2E,
	0x88, 0xF5, 0x75, 0xDA, 0xFF, 0xD8, 0x71, 0x80,
	0xA9, 0xDA, 0xF1, 0xFF, 0xD8, 0x82, 0xA7, 0xF3,
	0xC1, 0xF2, 0x80, 0xC2, 0xF1, 0x97, 0x86, 0x49,
	0x2E, 0xA6, 0xD0, 0x50, 0x96, 0x86, 0xAF, 0x75,
	0xD9, 0x88, 0xA2, 0xD0, 0xF3, 0xC0, 0xC3, 0xF1,
	0xDA, 0x8F, 0x96, 0xA2, 0xD0, 0xF3, 0xC2, 0xC3,
	0x82, 0xB6, 0x9B, 0x78, 0x78, 0xF1, 0xD8, 0x80,
	0xB7, 0x90, 0xAF, 0x0D, 0x89, 0x99, 0xAF, 0x10,
	0x80, 0x9F, 0x21, 0xDA, 0x2E, 0xD8, 0x89, 0x99,
	0xAF, 0x31, 0x80, 0xDA, 0x2E, 0xD8, 0xAF, 0x82,
	0x92, 0xF3, 0x41, 0x80, 0xF1, 0xD9, 0x2E, 0xD8,
	0xAF, 0x82, 0xF3, 0x19, 0x80, 0xF1, 0xD9, 0x2E,
	0xD8, 0xF1, 0x89, 0x90, 0xAF, 0xD0, 0x09, 0x8F,
	0x99, 0xAF, 0x51, 0xDB, 0x89, 0x31, 0xF3, 0x82,
	0x92, 0x19, 0xF2, 0xB1, 0x8C, 0xB5, 0x9C, 0x71,
	0xD9, 0xF1, 0xDF, 0xF9, 0xF2, 0xB9, 0xAC, 0xD0,
	0xF8, 0xF8, 0xF3, 0xDF, 0xD8, 0xB3, 0xB7, 0xBB,
	0x82, 0xAC, 0xF3, 0xC0, 0xA2, 0x80, 0x22, 0xF1,
	0xA9, 0x22, 0x26, 0x9F, 0xAF, 0x29, 0xDA, 0xAC,
	0xDE, 0xFF, 0xD8, 0xA2, 0xF2, 0xDE, 0xF1, 0xA9,
	0xDF, 0xF3, 0xBA, 0xAB, 0xD0, 0xF8, 0xF9, 0xD1,
	0xD9, 0xFF, 0xD8, 0xF2, 0xBB, 0xAD, 0xD0, 0xDE,
	/* bank 27: 0x1B00 */
	0xF8, 0xF1, 0xB1, 0x84, 0xB6, 0x96, 0xBA, 0xA7,
	0xD0, 0x7E, 0xB7, 0x96, 0xA7, 0x01, 0xB2, 0x87,
	0x9D, 0x05, 0xDB, 0xB3, 0x8D, 0xB6, 0x97, 0x79,
	0xF3, 0xB1, 0x8C, 0x96, 0x49, 0xF2, 0xBB, 0xAD,
	0xD0, 0xF8, 0xD8, 0xF3, 0xB9, 0xAC, 0xD0, 0xF8,
	0xF9, 0xD1, 0xD9, 0xF2, 0xBB, 0xAD, 0xD0, 0xF8,
	0xD8, 0xB3, 0xB7, 0xBB, 0x97, 0x8C, 0xAF, 0xF3,
	0x79, 0xD9, 0xF4, 0x1B, 0x5E, 0xD8, 0xF1, 0xB1,
	0x82, 0xB9, 0xA2, 0xD0, 0xC2, 0xB3, 0xF2, 0xB9,
	0xA3, 0xFA, 0xF1, 0xBB, 0xAA, 0xD0, 0xF8, 0xB8,
	0xB0, 0xB4, 0xA7, 0x88, 0x9C, 0xF7, 0x72, 0xF9,
	0xF4, 0xDA, 0x44, 0xD8, 0x1B, 0x71, 0xD8, 0xF3,
	0xB3, 0xB7, 0xBB, 0xA7, 0xD0, 0xFA, 0x97, 0x8C,
	0xAF, 0x79, 0xDA, 0xF1, 0x87, 0x9A, 0xAA, 0xD0,
	0x70, 0xD8, 0xF2, 0xBB, 0xB3, 0xB7, 0x82, 0x92,
	0xAF, 0x31, 0xDA, 0xF4, 0x1B, 0xB2, 0xD8, 0xF1,
	0xDE, 0x8D, 0x9D, 0xCC, 0x8F, 0x96, 0xA6, 0x00,
	0xAC, 0x8C, 0x9C, 0x0C, 0x30, 0xBA, 0x8D, 0x9D,
	0xA7, 0x39, 0xDB, 0xF3, 0xB1, 0x8C, 0xB6, 0x96,
	0x49, 0xD9, 0xF1, 0x84, 0xB5, 0x94, 0xB9, 0xA4,
	0xD0, 0x5E, 0xF0, 0xB7, 0x9D, 0x38, 0xD8, 0xF1,
	0xBB, 0xAC, 0xDE, 0xD0, 0xDE, 0xAD, 0xD0, 0xDF,
	0xF1, 0xFF, 0xD8, 0xF3, 0xB9, 0xAC, 0xD0, 0xF8,
	0xF9, 0xD1, 0xD9, 0xF2, 0xBB, 0xA2, 0xFA, 0xF8,
	0xDA, 0xF2, 0xBB, 0xA2, 0xFA, 0xD8, 0xF2, 0xB3,
	0xB7, 0xBB, 0x80, 0x92, 0xAF, 0x49, 0xD1, 0xD9,
	0xF1, 0xB9, 0xAC, 0xDE, 0xAD, 0xDE, 0xDF, 0xD8,
	0xF1, 0x8C, 0x9C, 0xBB, 0xAC, 0xD0, 0x10, 0xAC,
	0xDE, 0xAD, 0xD0, 0xDF, 0x92, 0x82, 0xAF, 0xF1,
	0xCA, 0xF2, 0x35, 0xF1, 0x96, 0x8F, 0xA6, 0xD9,
	0x00, 0xD8, 0xF1, 0xFF
};

struct nvi_dmp nvi_dmp_icm = {
	.fw				= dmp_fw_20628,
	.fw_len				= ARRAY_SIZE(dmp_fw_20628),
	.fw_crc32			= 0x12F362A6,
	.fw_mem_addr			= 0x90,
	.fw_start			= 0x08D0,
	.dmp_period_us			= DMP_PERIOD_US,
	.dev_msk			= (1 << DEV_SM) | (1 << DEV_QTN),
	.en_msk				= (1 << DEV_SM) | (1 << DEV_QTN),
	.dd_n				= ARRAY_SIZE(nvi_dmp_devs),
	.dd				= nvi_dmp_devs,
	.fn_init			= &nvi_dmp_init,
	.fn_en				= &nvi_dmp_en,
	.fn_dev_init			= &nvi_dd_init,
	.fn_dev_enable			= &nvi_dd_enable,
	.fn_dev_batch			= &nvi_dd_batch,
};
EXPORT_SYMBOL(nvi_dmp_icm);

