/* Copyright (c) 2016, NVIDIA CORPORATION.  All rights reserved.
 *
 * This software is licensed under the terms of the GNU General Public
 * License version 2, as published by the Free Software Foundation, and
 * may be copied, distributed, and modified under those terms.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#include <linux/delay.h>
#include "nvi.h"
#include "nvi_dmp_mpu.h"

#define DMP_PERIOD_US			(5000)


static int nvi_dmp_press_init(struct nvi_state *st)
{
	return 1;
}

static int nvi_dmp_cpass_init(struct nvi_state *st)
{
	int ret = 1;

	if (st->aux.port[0].nmp.type == SECONDARY_SLAVE_TYPE_COMPASS) {
		if (st->snsr[DEV_AUX].enable & (1 << 0))
			ret = nvi_aux_delay(st, __func__);
	}
	return ret;
}

static int nvi_dmp_sm_init(struct nvi_state *st)
{
	int ret;

	ret = nvi_mem_wr_be(st, D_SMD_MOT_THLD, 4,
			    st->snsr[DEV_SM].cfg.thresh_lo << 16);
	ret |= nvi_mem_wr_be(st, D_SMD_DELAY_THLD, 4,
			     st->snsr[DEV_SM].cfg.thresh_hi);
	ret |= nvi_mem_wr_be(st, D_SMD_DELAY2_THLD, 4,
			     st->snsr[DEV_SM].cfg.delay_us_max);
	return ret;
}

static struct nvi_dmp_dev nvi_dmp_devs[] = {
	{
		.dev			= DEV_ACC,
		.data_n			= 6,
		.hdr_n			= 2,
		.hdr			= { 0x40, 0x00 },
		.hdr_msk		= { 0xFF, 0xF0 },
		.mpu			= {
			.en_addr	= CFG_OUT_ACCL,
			.en_len		= 2,
			.en		= { 0xA3, 0xA3 },
			.dis		= { 0xF4, 0x12 },
			.odr_cfg	= KEY_CFG_ACCL_ODR,
			.odr_cntr	= KEY_ODR_CNTR_ACCL,
		},
	},
	{
		.dev			= DEV_GYR,
		.data_n			= 6,
		.hdr_n			= 2,
		.hdr			= { 0x20, 0x00 },
		.hdr_msk		= { 0xFF, 0xF0 },
		.mpu			= {
			.en_addr	= CFG_OUT_GYRO,
			.en_len		= 2,
			.en		= { 0xA3, 0xA3 },
			.dis		= { 0xF4, 0x12 },
			.odr_cfg	= KEY_CFG_GYRO_ODR,
			.odr_cntr	= KEY_ODR_CNTR_GYRO,
		},
	},
	{
		.dev			= DEV_QTN,
		.data_n			= 12,
		.hdr_n			= 4,
		.hdr			= { 0x04, 0x00, 0x04, 0x00 },
		.hdr_msk		= { 0xFF, 0xF0, 0xFF, 0xF0 },
		.mpu			= {
			.en_addr	= CFG_OUT_6QUAT,
			.en_len		= 2,
			.en		= { 0xA3, 0xA3 },
			.dis		= { 0xF4, 0x13 },
			.odr_cfg	= KEY_CFG_6QUAT_ODR,
			.odr_cntr	= KEY_ODR_CNTR_6QUAT,
		},
	},
	{
		.dev			= DEV_SM,
		.data_n			= 1,
		.hdr_n			= 0,
		.fn_init		= &nvi_dmp_sm_init,
		.mpu			= {
			.en_addr	= D_SMD_ENABLE,
			.en_len		= 2,
			.en		= { 0x00, 0x01 },
			.dis		= { 0x00, 0x00 },
			.odr_cfg	= KEY_CFG_ACCL_ODR,
			.odr_cntr	= KEY_ODR_CNTR_ACCL,
		},
	},
	{
		.dev			= DEV_AUX,
		.data_n			= 6,
		.aux_port		= 0,
		.hdr_n			= 2,
		.hdr			= { 0x10, 0x00 },
		.hdr_msk		= { 0xFF, 0xF0 },
		.fn_init		= &nvi_dmp_cpass_init,
		.mpu			= {
			.en_addr	= CFG_OUT_CPASS,
			.en_len		= 2,
			.en		= { 0xA3, 0xA3 },
			.dis		= { 0xF4, 0x12 },
			.odr_cfg	= KEY_CFG_CPASS_ODR,
			.odr_cntr	= KEY_ODR_CNTR_CPASS,
		},
	},
	{
		.dev			= DEV_AUX,
		.data_n			= 6,
		.aux_port		= 2,
		.hdr_n			= 2,
		.hdr			= { 0x80, 0x00 },
		.hdr_msk		= { 0xFF, 0xF0 },
		.fn_init		= &nvi_dmp_press_init,
		.mpu			= {
			.en_addr	= CFG_OUT_PRESS,
			.en_len		= 2,
			.en		= { 0xA3, 0xA3 },
			.dis		= { 0xF4, 0x12 },
			.odr_cfg	= KEY_CFG_PRESS_ODR,
			.odr_cntr	= KEY_ODR_CNTR_PRESS,
		},
	},
};

static int nvi_dd_odr(struct nvi_state *st, struct nvi_dmp_dev *dd)
{
	u16 odr_cfg;
	unsigned int period_us;
	int ret;

	if (dd->dev == DEV_AUX)
		period_us = st->aux.port[dd->aux_port].period_us;
	else
		period_us = st->snsr[dd->dev].period_us;
	if (period_us)
		odr_cfg = (period_us / DMP_PERIOD_US) - 1;
	else
		odr_cfg = 1;
	ret = nvi_mem_wr_be(st, dd->mpu.odr_cntr, 2, 0);
	ret |= nvi_mem_wr_be(st, dd->mpu.odr_cfg, 2, odr_cfg);
	ret |= nvi_mem_wr_be(st, dd->mpu.odr_cntr, 2, 0);
	return ret;
}

static int nvi_dd_en(struct nvi_state *st, struct nvi_dmp_dev *dd)
{
	int ret;

	if (st->snsr[dd->dev].enable) {
		if (dd->fn_init)
			ret = dd->fn_init(st);
		else
			ret = 0;
		if (ret > 0) {
			/* ignore without error */
			ret = 0;
		} else if (!ret) {
			ret = nvi_dd_odr(st, dd);
			ret |= nvi_mem_wr(st, dd->mpu.en_addr, dd->mpu.en_len,
					  &dd->mpu.en[0], false);
		}
		if (ret && (dd->dev != DEV_AUX))
			st->en_msk &= ~(1 << dd->dev);
	} else {
		ret = nvi_mem_wr(st, dd->mpu.en_addr, dd->mpu.en_len,
				 &dd->mpu.dis[0], false);
	}
	return ret;
}

static struct nvi_dmp_dev *nvi_dd(struct nvi_state *st,
				  unsigned int dev, int port)
{
	unsigned int i;

	for (i = 0; i < ARRAY_SIZE(nvi_dmp_devs); i++) {
		if (dev == nvi_dmp_devs[i].dev) {
			if (dev == DEV_AUX) {
				if (port != nvi_dmp_devs[i].aux_port)
					continue;
			}
			return &nvi_dmp_devs[i];
		}
	}

	return NULL;
}

static int nvi_dd_batch(struct nvi_state *st, unsigned int dev, int port)
{
	struct nvi_dmp_dev *dd;

	dd = nvi_dd(st, dev, port);
	if (dd == NULL)
		return -EINVAL;

	return nvi_dd_odr(st, dd);
}

static int nvi_dd_enable(struct nvi_state *st, unsigned int dev, int port)
{
	struct nvi_dmp_dev *dd;

	dd = nvi_dd(st, dev, port);
	if (dd == NULL)
		return -EINVAL;

	return nvi_dd_en(st, dd);
}

static int nvi_dd_init(struct nvi_state *st, unsigned int dev)
{
	struct nvi_dmp_dev *dd;
	int ret = -EINVAL;

	dd = nvi_dd(st, dev, -1);
	if (dd == NULL)
		return -EINVAL;

	if (dd->fn_init)
		ret = dd->fn_init(st);
	return ret;
}


static u8 nvi_dmp_fcfg_gyro_a[8][3] = {
	{ 0x4C, 0xCD, 0x6C },
	{ 0xCD, 0x4C, 0x6C },
	{ 0x4C, 0xCD, 0x6C },
	{ 0xCD, 0x4C, 0x6C },
	{ 0x4C, 0xCD, 0x6C },
	{ 0xCD, 0x4C, 0x6C },
	{ 0x4C, 0xCD, 0x6C },
	{ 0xCD, 0x4C, 0x6C },
};

static u8 nvi_dmp_fcfg_gyro_b[8][3] = {
	{ 0x36, 0x56, 0x76 },
	{ 0x36, 0x57, 0x76 },
	{ 0x37, 0x57, 0x76 },
	{ 0x37, 0x56, 0x76 },
	{ 0x37, 0x46, 0x77 },
	{ 0x37, 0x56, 0x77 },
	{ 0x36, 0x57, 0x77 },
	{ 0x36, 0x56, 0x77 },
};

static u8 nvi_dmp_fcfg_accl_a[8][3] = {
	{ 0x0C, 0xC9, 0x2C },
	{ 0xC9, 0x0C, 0x2C },
	{ 0x0C, 0xC9, 0x2C },
	{ 0xC9, 0x0C, 0x2C },
	{ 0x0C, 0xC9, 0x2C },
	{ 0xC9, 0x0C, 0x2C },
	{ 0x0C, 0xC9, 0x2C },
	{ 0xC9, 0x0C, 0x2C },
};

static u8 nvi_dmp_fcfg_accl_b[8][3] = {
	{ 0x26, 0x46, 0x66 },
	{ 0x26, 0x47, 0x66 },
	{ 0x27, 0x47, 0x66 },
	{ 0x27, 0x46, 0x66 },
	{ 0x27, 0x46, 0x67 },
	{ 0x27, 0x46, 0x66 },
	{ 0x26, 0x47, 0x67 },
	{ 0x26, 0x46, 0x67 },
};

static signed char nvi_matrix[8][9] = {
	{  1,  0,  0,  0,  1,  0,  0,  0,  1 },
	{  0,  1,  0, -1,  0,  0,  0,  0,  1 },
	{ -1,  0,  0,  0, -1,  0,  0,  0,  1 },
	{  0, -1,  0,  1,  0,  0,  0,  0,  1 },
	{ -1,  0,  0,  0,  1,  0,  0,  0, -1 },
	{  0, -1,  0,  1,  0,  0,  0,  0, -1 },
	{  1,  0,  0,  0, -1,  0,  0,  0, -1 },
	{  0,  1,  0,  1,  0,  0,  0,  0, -1 },
};

static int nvi_dmp_matrix_i(signed char *matrix)
{
	int i;
	int j;

	for (i = 0; i < 8; i++) {
		for (j = 0; j < 9; j++) {
			if (matrix[j] != nvi_matrix[i][j])
				break;
		}
		if (j >= 9)
			return i;
	}

	pr_err("%s ERR: matrix=%d,%d,%d,%d,%d,%d,%d,%d,%d\n",
	       __func__, matrix[0], matrix[1], matrix[2], matrix[3],
	       matrix[4], matrix[5], matrix[6], matrix[7], matrix[8]);
	return -EINVAL;
}

static int nvi_dmp_init(struct nvi_state *st)
{
	int i;
	int ret;

	i = nvi_dmp_matrix_i(st->snsr[DEV_GYR].cfg.matrix);
	if (i < 0)
		return -EINVAL;

	ret = nvi_mem_wr(st, FCFG_1, 3, &nvi_dmp_fcfg_gyro_a[i][0], false);
	ret |= nvi_mem_wr(st, FCFG_3, 3, &nvi_dmp_fcfg_gyro_b[i][0], false);
	i = nvi_dmp_matrix_i(st->snsr[DEV_ACC].cfg.matrix);
	if (i < 0)
		return -EINVAL;

	ret |= nvi_mem_wr(st, FCFG_2, 3, &nvi_dmp_fcfg_accl_a[i][0], false);
	ret |= nvi_mem_wr(st, FCFG_7, 3, &nvi_dmp_fcfg_accl_b[i][0], false);
	ret |= nvi_pm_wr(st, __func__, INV_CLK_PLL, 0, 0);
	st->en_msk |= (1 << DEV_DMP);
	ret |= nvi_user_ctrl_en(st, __func__, true, false, false, false);
	st->en_msk &= ~(1 << DEV_DMP);
	if (ret)
		return ret;

	msleep(400);
	ret |= nvi_user_ctrl_en(st, __func__, false, false, false, false);
	return ret;
}

static int nvi_dmp_en(struct nvi_state *st)
{
	unsigned int i;
	int ret = 0;

	ret |= nvi_i2c_wr_rc(st, &st->hal->reg->gyro_config1, 0x03,
			     __func__, &st->rc.gyro_config1);
	ret |= nvi_i2c_wr_rc(st, &st->hal->reg->gyro_config2, 0x18,
			     __func__, &st->rc.gyro_config2);
	ret |= nvi_i2c_wr_rc(st, &st->hal->reg->accel_config, 0,
			     __func__, &st->rc.accel_config);
	ret |= nvi_i2c_wr_rc(st, &st->hal->reg->smplrt[0], 0x04,
			     __func__, (u8 *)&st->rc.smplrt[0]);
	for (i = 0; i < ARRAY_SIZE(nvi_dmp_devs); i++)
		ret |= nvi_dd_en(st, &nvi_dmp_devs[i]);
	return ret;
}

static const u8 const dmp_fw_6515[] = {
	/* bank 0: 0x0020 */
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x40, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x01, 0x00, 0x7D, 0x00, 0x01,
	0x00, 0x0F, 0x00, 0x00, 0x00, 0x00, 0x00, 0x09,
	0x00, 0x83, 0x12, 0x6F, 0xFF, 0xDF, 0x3B, 0x64,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x3E, 0x03, 0x30, 0x40, 0x00, 0x00, 0x00,
	0x02, 0xCA, 0xE3, 0x09, 0x3E, 0x80, 0x00, 0x00,
	0x20, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x40, 0x00, 0x00, 0x00, 0x60, 0x00, 0x00, 0x00,
	0x00, 0x0C, 0x00, 0x00, 0x00, 0x0C, 0x18, 0x6E,
	0x00, 0x00, 0x06, 0x92, 0x0A, 0x16, 0xC0, 0xDF,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x02,
	0x00, 0x00, 0x00, 0x04, 0x00, 0x00, 0x00, 0x07,
	0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0x0B, 0x2B,
	0x00, 0x00, 0x16, 0x57, 0x00, 0x00, 0x03, 0x59,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x1D, 0xFA,
	0x00, 0x02, 0x6C, 0x1D, 0x00, 0x00, 0x00, 0x00,
	0x40, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x07, 0x00, 0x00, 0x00, 0x0E, 0x00, 0x00,
	0x00, 0x09, 0x00, 0x00, 0x00, 0x10, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x01, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	/* bank 1: 0x0100 */
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x28, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x10, 0x72, 0x00, 0x00, 0x00, 0x40,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x02, 0x58,
	0x05, 0xDC, 0x00, 0x00, 0x00, 0x00, 0x00, 0xC8,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x03, 0x00, 0x00, 0x00, 0xFF,
	0x08, 0x00, 0x02, 0x00, 0x04, 0x00, 0x01, 0x00,
	0x80, 0x00, 0x10, 0x00, 0x20, 0x00, 0x40, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0xB2, 0x6A, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	/* bank 2: 0x0200 */
	0x00, 0x28, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x0B, 0xEB, 0xC2, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x14,
	0x00, 0x00, 0x00, 0x00, 0x02, 0x03, 0x18, 0x85,
	0x04, 0x06, 0x31, 0x0A, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x1B, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x05, 0x00, 0x05, 0x00, 0x05, 0x00, 0x05,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x1B, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x05, 0x00, 0x05, 0x00, 0x05, 0x00, 0x05,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x05, 0x00, 0x05, 0x00, 0x05, 0x00, 0x05,
	0xFF, 0xE5, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	/* bank 3: 0x0300 */
	0x00, 0x00, 0x00, 0x01, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x01, 0x80, 0x00, 0x00, 0x01, 0x80, 0x00,
	0x00, 0x01, 0x80, 0x00, 0x00, 0x24, 0x26, 0xD3,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x07, 0x00, 0x10, 0x00, 0x96, 0x00, 0x3C,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x0C, 0x0A, 0x4E, 0x68, 0xCD, 0xCF, 0x77, 0x09,
	0x50, 0x16, 0x67, 0x59, 0xC6, 0x19, 0xCE, 0x82,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x10, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x08,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x10, 0x00, 0x00, 0x00, 0x3C,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x02, 0x03, 0x18, 0x85, 0x00, 0x00, 0x40, 0x00,
	0x00, 0x00, 0x00, 0x03, 0x00, 0x00, 0x00, 0x03,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x40, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x02,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x67, 0x7D, 0xDF, 0x7E,
	0x72, 0x90, 0x2E, 0x55, 0x4C, 0xF6, 0xE6, 0x88,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	/* bank 4: 0x0400 */
	0xD8, 0xB0, 0xB5, 0xB9, 0xF3, 0xA6, 0xF8, 0xF9,
	0xD1, 0xD9, 0x81, 0x96, 0xF8, 0xF7, 0x3E, 0xD8,
	0xF3, 0xB1, 0x86, 0x96, 0xA3, 0x31, 0xD1, 0xDA,
	0xF1, 0xFF, 0xD8, 0xB3, 0xB7, 0xBB, 0x8E, 0x9E,
	0xAE, 0xF1, 0x32, 0xF5, 0x1B, 0xF1, 0xB4, 0xB8,
	0xB0, 0x80, 0x97, 0xF1, 0xA9, 0xDF, 0xDF, 0xDF,
	0xAA, 0xDF, 0xDF, 0xDF, 0xF2, 0xAA, 0xC5, 0xCD,
	0xC7, 0xA9, 0x0C, 0xC9, 0x2C, 0x97, 0xF1, 0xA9,
	0x89, 0x26, 0x46, 0x66, 0xB2, 0x89, 0x99, 0xA9,
	0x2D, 0x55, 0x7D, 0xB0, 0x8A, 0xA8, 0x96, 0x36,
	0x56, 0x76, 0xD8, 0xF1, 0xB0, 0x89, 0xB9, 0xA3,
	0xC3, 0xC5, 0xC7, 0xB1, 0x81, 0xB4, 0x97, 0xA3,
	0x11, 0xB5, 0x93, 0xA1, 0xF0, 0x24, 0x08, 0x44,
	0x10, 0x64, 0x18, 0xF1, 0x93, 0x81, 0xA3, 0x2D,
	0x55, 0x3D, 0xF2, 0xA6, 0xFA, 0xF9, 0xD1, 0xD9,
	0xF8, 0xF1, 0x83, 0xA2, 0xC3, 0xC5, 0xC7, 0xD8,
	0xF3, 0xA2, 0xDE, 0xF1, 0x82, 0x93, 0xA3, 0x2D,
	0x55, 0x7D, 0x83, 0x95, 0xF5, 0xA3, 0x30, 0xD9,
	0xF3, 0xA2, 0xF8, 0xD8, 0xF5, 0xA3, 0x50, 0xD9,
	0xF3, 0xA2, 0xF8, 0xD8, 0xF5, 0xA3, 0x70, 0xD9,
	0xF3, 0xA2, 0xF8, 0xD8, 0xF2, 0xB9, 0xA2, 0xF8,
	0xF9, 0xD1, 0xD9, 0xA6, 0xDE, 0xDF, 0xF4, 0x1C,
	0xD8, 0xF2, 0xA6, 0xF8, 0xF9, 0xD1, 0xD9, 0xF4,
	0x14, 0xD8, 0xF2, 0xF9, 0xD1, 0xD9, 0xF8, 0xF4,
	0x1E, 0xD8, 0xF2, 0xF9, 0xD1, 0xD9, 0xF8, 0xF8,
	0xF4, 0x14, 0xD8, 0xF4, 0x10, 0xD8, 0xF3, 0xA2,
	0xF8, 0xF9, 0xD1, 0xDA, 0xF2, 0xA6, 0xF8, 0xF1,
	0xA5, 0xDE, 0xD8, 0xF4, 0xA3, 0x14, 0x14, 0xD8,
	0xF1, 0xA5, 0xF8, 0xA3, 0x85, 0x95, 0x09, 0xD9,
	0xF1, 0xA5, 0xDE, 0xF2, 0xA6, 0xF8, 0xDF, 0xD8,
	0xF4, 0xA3, 0x09, 0xD8, 0xF3, 0xA2, 0xF8, 0xF9,
	0xD1, 0xF4, 0xD9, 0x08, 0x17, 0xDA, 0x42, 0xF2,
	/* bank 5: 0x0500 */
	0xA2, 0xDE, 0xF4, 0x0B, 0xD8, 0xF1, 0xA5, 0xF8,
	0xA3, 0x85, 0x95, 0x19, 0xDA, 0xF4, 0x05, 0xD8,
	0xF2, 0xA6, 0xDE, 0xDF, 0xD8, 0xF1, 0xB8, 0xAA,
	0xB3, 0x8D, 0xB4, 0x98, 0x0D, 0x35, 0x5D, 0xB2,
	0xB6, 0xBA, 0xAF, 0x8C, 0x96, 0x19, 0x8F, 0x9F,
	0xB8, 0xA8, 0x0E, 0x16, 0x1E, 0xB4, 0x9A, 0xAA,
	0x88, 0x2C, 0x54, 0x7C, 0xD8, 0xF1, 0xB8, 0xB4,
	0xB0, 0x97, 0x86, 0xA8, 0x31, 0x9B, 0x06, 0x99,
	0x07, 0xAB, 0x97, 0x28, 0x88, 0x9B, 0xF0, 0x0C,
	0x20, 0x14, 0x40, 0xB0, 0xB4, 0xB8, 0xF0, 0xA8,
	0x8A, 0x9A, 0x28, 0x50, 0x78, 0xB7, 0x9B, 0xA8,
	0x29, 0x51, 0x79, 0x24, 0x70, 0x59, 0x44, 0x69,
	0x38, 0x64, 0x48, 0x31, 0xF1, 0xBB, 0xAB, 0x88,
	0x00, 0x2C, 0x54, 0x7C, 0xF0, 0xB3, 0x8B, 0xB8,
	0xA8, 0x04, 0x28, 0x50, 0x78, 0xF1, 0xB0, 0x88,
	0xB4, 0x97, 0x26, 0xA8, 0x59, 0x98, 0xBB, 0xAB,
	0xB3, 0x8B, 0x02, 0x26, 0x46, 0x66, 0xB0, 0xB8,
	0xF0, 0x8A, 0x9C, 0xA8, 0x29, 0x51, 0x79, 0x8B,
	0x29, 0x51, 0x8A, 0x24, 0x70, 0x59, 0x8B, 0x20,
	0x58, 0x8A, 0x44, 0x69, 0x38, 0x8B, 0x39, 0x40,
	0x8A, 0x64, 0x48, 0x31, 0x8B, 0x30, 0x49, 0x88,
	0xF1, 0xAC, 0x00, 0x2C, 0x54, 0x7C, 0xF0, 0x8C,
	0xA8, 0x04, 0x28, 0x50, 0x78, 0xF1, 0x88, 0x97,
	0x26, 0xA8, 0x59, 0x98, 0xAC, 0x8C, 0x02, 0x26,
	0x46, 0x66, 0xF0, 0x89, 0x9C, 0xA8, 0x29, 0x51,
	0x79, 0x24, 0x70, 0x59, 0x44, 0x69, 0x38, 0x64,
	0x48, 0x31, 0xA9, 0x88, 0x09, 0x20, 0x59, 0x70,
	0xAB, 0x11, 0x38, 0x40, 0x69, 0xA8, 0x19, 0x31,
	0x48, 0x60, 0x8C, 0xA8, 0x3C, 0x41, 0x5C, 0x20,
	0x7C, 0x00, 0xF1, 0x87, 0x98, 0x19, 0x86, 0xA8,
	0x6E, 0x76, 0x7E, 0xA9, 0x99, 0x88, 0x2D, 0x55,
	0x7D, 0xD8, 0xF1, 0xB3, 0x8B, 0xB4, 0x97, 0xBB,
	/* bank 6: 0x0600 */
	0xAB, 0xF8, 0xF9, 0xB9, 0xA3, 0xDA, 0xC3, 0xC5,
	0xC7, 0xD9, 0x2D, 0x4D, 0x6D, 0xD8, 0xB0, 0x8C,
	0xB8, 0xAC, 0xF8, 0xF9, 0xA8, 0xDA, 0xC3, 0xC5,
	0xC7, 0xD9, 0x2D, 0x4D, 0x6D, 0xD8, 0xF3, 0xB9,
	0xAC, 0xDE, 0xD8, 0xB1, 0xB6, 0xB9, 0xF1, 0xA8,
	0xF8, 0xF3, 0xB9, 0xAA, 0xFA, 0xF9, 0xD1, 0xDA,
	0xF2, 0x8A, 0xCA, 0xF4, 0x0E, 0xF3, 0xB1, 0x87,
	0xDD, 0xC2, 0xC2, 0xF1, 0x8C, 0xC4, 0xDC, 0xF3,
	0xB9, 0xAC, 0xF8, 0xD8, 0xF2, 0xB2, 0xB6, 0xBA,
	0xAD, 0xFA, 0x8D, 0x9D, 0xAB, 0x39, 0xD9, 0xAD,
	0xDF, 0xF4, 0x13, 0xDD, 0xF2, 0xB1, 0x87, 0xB5,
	0x9A, 0x08, 0x08, 0xF1, 0xB1, 0x83, 0xC2, 0xC4,
	0xC6, 0xDC, 0xF3, 0xB9, 0xAC, 0xF8, 0xD8, 0xF3,
	0xB2, 0xB6, 0xBA, 0xAD, 0xF8, 0x8D, 0x9D, 0xAB,
	0x11, 0xD9, 0xAD, 0xDE, 0xF4, 0x13, 0xDD, 0xF2,
	0xB1, 0x87, 0xB5, 0x9A, 0x28, 0x28, 0xF1, 0xB0,
	0x88, 0xC2, 0xC4, 0xC6, 0xDC, 0xF3, 0xB9, 0xAC,
	0xF8, 0xD8, 0xF3, 0xB2, 0xB6, 0xBA, 0xAD, 0xFA,
	0x8D, 0x9D, 0xAB, 0x39, 0xD9, 0xAD, 0xDF, 0xF4,
	0x12, 0xDD, 0xF3, 0xB1, 0x87, 0xB5, 0x9A, 0x08,
	0xF2, 0xF2, 0x88, 0xC2, 0xC4, 0xC6, 0xDC, 0xF3,
	0xB9, 0xAC, 0xF8, 0xD8, 0xF2, 0xB2, 0xB6, 0xBA,
	0xAE, 0xF8, 0x8E, 0x9E, 0xAB, 0x11, 0xD9, 0xAE,
	0xDE, 0xF4, 0x12, 0xDD, 0xF3, 0xB1, 0x87, 0xB5,
	0x9A, 0x68, 0xF2, 0xB0, 0x80, 0xC0, 0xC8, 0xC2,
	0xDC, 0xF3, 0xB9, 0xAC, 0xF8, 0xD8, 0xF2, 0xB2,
	0xB6, 0xBA, 0xAE, 0xFA, 0x8E, 0x9E, 0xAB, 0x39,
	0xD9, 0xAE, 0xDF, 0xF4, 0x12, 0xDD, 0xF2, 0xB1,
	0x87, 0xB5, 0x9A, 0x68, 0xF2, 0xB0, 0x80, 0xC4,
	0xCC, 0xC6, 0xDC, 0xF3, 0xB9, 0xAC, 0xF8, 0xD8,
	0xF3, 0xB2, 0xB6, 0xBA, 0xAE, 0xF8, 0x8E, 0x9E,
	0xAB, 0x11, 0xD9, 0xAE, 0xDE, 0xF4, 0x12, 0xDD,
	/* bank 7: 0x0700 */
	0xF3, 0xB1, 0x87, 0xB5, 0x9A, 0x48, 0xF2, 0xB0,
	0x81, 0xC0, 0xC8, 0xC2, 0xDC, 0xF3, 0xB9, 0xAC,
	0xF8, 0xD8, 0xF3, 0xB2, 0xB6, 0xBA, 0xAE, 0xFA,
	0x8E, 0x9E, 0xAB, 0x39, 0xD9, 0xAE, 0xDF, 0xF4,
	0x12, 0xDD, 0xF2, 0xB1, 0x87, 0xB5, 0x9A, 0x48,
	0xF2, 0xB0, 0x81, 0xC4, 0xCC, 0xC6, 0xDC, 0xF3,
	0xB9, 0xAC, 0xF8, 0xD8, 0xF2, 0xB0, 0x88, 0xB9,
	0xA8, 0xC3, 0xC5, 0xC7, 0xB1, 0xB5, 0xB9, 0xF3,
	0xAC, 0xFA, 0xF9, 0xD1, 0xDA, 0xF4, 0x10, 0xD8,
	0xF3, 0xAC, 0xF8, 0xF9, 0xD1, 0xDA, 0xF1, 0xB9,
	0xAA, 0xDF, 0xF4, 0x1C, 0xD9, 0xF4, 0x1A, 0xD8,
	0xF3, 0xAC, 0xF8, 0xF9, 0xD1, 0xD9, 0xF4, 0x1D,
	0xD8, 0xF2, 0x8C, 0xAC, 0xF8, 0xF9, 0xD1, 0xD9,
	0xC2, 0xD8, 0xF2, 0xF9, 0xD9, 0xDE, 0xF4, 0x09,
	0x1D, 0x16, 0xDA, 0xF2, 0xDD, 0xC6, 0xDC, 0xF4,
	0x30, 0xF1, 0xB9, 0xAA, 0xDF, 0xD8, 0xF1, 0xAB,
	0xFA, 0x8B, 0x9B, 0xA3, 0x69, 0xD9, 0xF4, 0x07,
	0x06, 0xDA, 0xF1, 0xB9, 0xAB, 0xDF, 0xFE, 0xD8,
	0xF1, 0xBB, 0xB3, 0xB7, 0xAA, 0xF9, 0xDA, 0xFF,
	0xD9, 0x80, 0x9A, 0xAA, 0x28, 0xB4, 0x80, 0x98,
	0xA7, 0x20, 0xD8, 0xF1, 0xB3, 0xB7, 0xBB, 0x87,
	0xAA, 0xD0, 0xC1, 0xA7, 0x81, 0x97, 0x62, 0x93,
	0xF0, 0x71, 0x71, 0x60, 0x85, 0x94, 0x01, 0x29,
	0x51, 0x79, 0x90, 0xA5, 0xF1, 0x28, 0x4C, 0x6C,
	0x87, 0x0C, 0x95, 0x18, 0x85, 0x78, 0xA3, 0x83,
	0x90, 0x28, 0x4C, 0x6C, 0x8A, 0x6C, 0xF2, 0xB0,
	0xB4, 0xB8, 0x83, 0x93, 0xA5, 0x71, 0xDB, 0xF1,
	0x8C, 0x01, 0xA3, 0xC0, 0xD8, 0xF2, 0x83, 0xA5,
	0x71, 0xD9, 0xA3, 0xD0, 0xF8, 0xF1, 0x9C, 0xA5,
	0xDB, 0x21, 0x8C, 0xA3, 0xC1, 0xD8, 0xF2, 0x83,
	0xA5, 0xC7, 0xF9, 0xF1, 0x93, 0x25, 0xF2, 0xA5,
	0x95, 0x49, 0xD9, 0xF3, 0xA3, 0xD0, 0xDE, 0xF8,
	/* bank 8: 0x0800 */
	0xF1, 0x84, 0xA5, 0xD0, 0xDB, 0x55, 0xF3, 0xA3,
	0xD0, 0xDE, 0xD8, 0xF2, 0xA5, 0x83, 0x95, 0x49,
	0xDB, 0xF3, 0x8E, 0x93, 0x11, 0xF1, 0x85, 0x94,
	0xA5, 0x79, 0x9E, 0x83, 0xF3, 0x69, 0xA3, 0xD0,
	0xF8, 0xD8, 0xF2, 0xA5, 0x83, 0x95, 0x49, 0xD9,
	0xF3, 0xA3, 0xD0, 0xC5, 0xF2, 0xA3, 0xD0, 0xDE,
	0xF1, 0xA3, 0xDE, 0x87, 0xC5, 0xD8, 0xF3, 0x8E,
	0x93, 0xA5, 0x11, 0xDB, 0xF2, 0x8D, 0xB7, 0x92,
	0x31, 0xBB, 0xA2, 0xD0, 0xC2, 0xA8, 0xB3, 0x82,
	0xB4, 0x9D, 0x21, 0xB0, 0x8D, 0xB7, 0x92, 0x49,
	0xA2, 0xDF, 0xD8, 0xF3, 0xB4, 0x9E, 0xB0, 0x83,
	0xB8, 0xA5, 0x49, 0xDB, 0xF2, 0xB3, 0x82, 0x9D,
	0x41, 0x21, 0xB0, 0x84, 0xB7, 0x92, 0x09, 0xD9,
	0xB3, 0x82, 0xBB, 0xA8, 0xF1, 0xCA, 0x96, 0x88,
	0xA6, 0x00, 0xD8, 0xF3, 0xB7, 0x90, 0xB0, 0x83,
	0xB8, 0xA5, 0x49, 0xDB, 0xF2, 0xB3, 0x82, 0xB4,
	0x9D, 0x41, 0xB0, 0x8D, 0xBB, 0xA2, 0xD0, 0xC0,
	0xB3, 0x82, 0xB4, 0x9D, 0xA8, 0x21, 0xA2, 0xDF,
	0xD8, 0xF3, 0xBB, 0xB3, 0xB7, 0x90, 0xA2, 0x82,
	0x00, 0xF2, 0x10, 0xA8, 0x92, 0x19, 0x80, 0xA2,
	0xF2, 0xD9, 0x26, 0xF3, 0xA7, 0xD0, 0xDF, 0xD8,
	0xF1, 0xA8, 0x89, 0x95, 0x19, 0xA9, 0x80, 0xD9,
	0x38, 0xD8, 0xA8, 0x89, 0x39, 0xA9, 0x80, 0xDA,
	0x3C, 0xD8, 0xA8, 0x2E, 0x85, 0xF5, 0x75, 0xDA,
	0xFF, 0xD8, 0x71, 0x80, 0xA9, 0xDA, 0xF1, 0xFF,
	0xD8, 0xD8, 0xF1, 0xB2, 0xB7, 0xBA, 0x83, 0x99,
	0xA4, 0x09, 0xDA, 0xA3, 0xDF, 0xF9, 0xC7, 0xF4,
	0x0B, 0xD8, 0xF1, 0xA3, 0xFB, 0xF8, 0xDB, 0xFB,
	0xD9, 0xDE, 0xA3, 0xDF, 0xD8, 0xF1, 0xA3, 0xFA,
	0xF9, 0xBB, 0xA9, 0xD0, 0x84, 0xDA, 0xC2, 0xD9,
	0xC4, 0xD8, 0xB3, 0x82, 0xA7, 0xF3, 0xC1, 0xF2,
	0x80, 0xC2, 0xF1, 0x97, 0x86, 0x49, 0x2E, 0xA6,
	/* bank 9: 0x0900 */
	0xD0, 0x50, 0x96, 0x86, 0xA8, 0x75, 0xD9, 0x88,
	0xA2, 0xD0, 0xF3, 0xC4, 0xC7, 0xF1, 0xDA, 0x88,
	0x96, 0xA2, 0xD0, 0xF3, 0xC2, 0xC3, 0x82, 0xB4,
	0x94, 0x68, 0x68, 0xF1, 0xD8, 0x80, 0xB7, 0x90,
	0xA8, 0x0D, 0x89, 0x99, 0xA8, 0x10, 0x80, 0x98,
	0x21, 0xDA, 0x2E, 0xD8, 0x89, 0x99, 0xA8, 0x31,
	0x80, 0xDA, 0x2E, 0xD8, 0xA8, 0x82, 0x92, 0xF3,
	0x41, 0x80, 0xF1, 0xD9, 0x2E, 0xD8, 0xA8, 0x82,
	0xF3, 0x19, 0x80, 0xF1, 0xD9, 0x2E, 0xD8, 0x82,
	0xAC, 0xF3, 0xC0, 0xA2, 0x80, 0x22, 0xF1, 0xA9,
	0x22, 0x26, 0x98, 0xA8, 0x29, 0xDA, 0xAC, 0xDE,
	0xFF, 0xD8, 0xA2, 0xF2, 0xDE, 0xF1, 0xA9, 0xDF,
	0x97, 0x8C, 0xA8, 0xF3, 0x79, 0xDA, 0xF1, 0xF1,
	0xF1, 0xF1, 0xB1, 0x88, 0xB9, 0xAC, 0xD0, 0xC0,
	0xB3, 0xF3, 0xF3, 0xB9, 0xAA, 0xFA, 0xF1, 0xBB,
	0xAA, 0xD0, 0xF8, 0xF4, 0x10, 0xD8, 0xF3, 0xA7,
	0xD0, 0xFA, 0x97, 0x8C, 0xA8, 0x79, 0xDA, 0xF1,
	0x87, 0x9A, 0xAA, 0xD0, 0x70, 0xD8, 0xF2, 0x82,
	0x92, 0xA8, 0x31, 0x80, 0xA6, 0x96, 0xF1, 0xD9,
	0x00, 0xAC, 0x8C, 0x9C, 0x0C, 0x30, 0xAC, 0xDE,
	0xD0, 0xDE, 0xF1, 0xFF, 0xD8, 0x8C, 0x9C, 0xAC,
	0xD0, 0x10, 0xAC, 0xDE, 0x80, 0x92, 0xA2, 0xF2,
	0x4C, 0x82, 0xA8, 0xF1, 0xCA, 0xF2, 0x35, 0xF1,
	0x96, 0x88, 0xA6, 0xD9, 0x00, 0xD8, 0xFF
};

struct nvi_dmp nvi_dmp_mpu = {
	.fw				= dmp_fw_6515,
	.fw_len				= ARRAY_SIZE(dmp_fw_6515),
	.fw_crc32			= 0x972AAE92,
	.fw_mem_addr			= 0x20,
	.fw_start			= 0x0400,
	.dmp_period_us			= DMP_PERIOD_US,
	.dev_msk			= (1 << DEV_SM) | (1 << DEV_QTN),
	.en_msk				= (1 << DEV_SM) | (1 << DEV_QTN),
	.dd_n				= ARRAY_SIZE(nvi_dmp_devs),
	.dd				= nvi_dmp_devs,
	.fn_init			= &nvi_dmp_init,
	.fn_en				= &nvi_dmp_en,
	.fn_dev_init			= &nvi_dd_init,
	.fn_dev_enable			= &nvi_dd_enable,
	.fn_dev_batch			= &nvi_dd_batch,
};
EXPORT_SYMBOL(nvi_dmp_mpu);

