/*
 * drivers/video/tegra/host/vi/vi.c
 *
 * Tegra Graphics Host VI
 *
 * Copyright (c) 2012-2015, NVIDIA CORPORATION. All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <linux/init.h>
#include <linux/export.h>
#include <linux/module.h>
#include <linux/resource.h>
#include <linux/pm_runtime.h>
#include <linux/of.h>
#include <linux/of_device.h>
#include <linux/of_platform.h>
#include <linux/clk/tegra.h>
#include <linux/tegra-soc.h>
#include <linux/tegra_pm_domains.h>
#include <linux/debugfs.h>

#include <media/tegra_v4l2_camera.h>

#include "dev.h"
#include "bus_client.h"
#include "nvhost_acm.h"
#include "t124/t124.h"
#include "t210/t210.h"
#include "vi.h"
#include "vi_irq.h"
#include "vi_common.h"
#include "camera_priv_defs.h"

#ifdef CONFIG_ARCH_TEGRA_18x_SOC
#include "t186/t186.h"
#endif

#define MAX_DEVID_LENGTH	16
#define TEGRA_VI_NAME		"tegra_vi"

static struct of_device_id tegra_vi_of_match[] = {
#ifdef TEGRA_12X_OR_HIGHER_CONFIG
	{ .compatible = "nvidia,tegra124-vi",
		.data = (struct nvhost_device_data *)&t124_vi_info },
#endif
#ifdef TEGRA_21X_OR_HIGHER_CONFIG
	{ .compatible = "nvidia,tegra210-vi",
		.data = (struct nvhost_device_data *)&t21_vi_info },
#endif
#ifdef CONFIG_ARCH_TEGRA_18x_SOC
	{ .compatible = "nvidia,tegra186-vi",
		.data = (struct nvhost_device_data *)&t18_vi_info },
#endif
	{ },
};

static struct i2c_camera_ctrl *i2c_ctrl;

static void (*mfi_callback)(void *);
static void *mfi_callback_arg;
static DEFINE_MUTEX(vi_isr_lock);

#ifdef CONFIG_VIDEO_TEGRA_VI
int tegra_vi_register_mfi_cb(callback cb, void *cb_arg)
{
	if (mfi_callback || mfi_callback_arg) {
		pr_err("cb already registered\n");
		return -1;
	}

	mutex_lock(&vi_isr_lock);
	mfi_callback = cb;
	mfi_callback_arg = cb_arg;
	mutex_unlock(&vi_isr_lock);

	return 0;
}
EXPORT_SYMBOL(tegra_vi_register_mfi_cb);

int tegra_vi_unregister_mfi_cb(void)
{
	mutex_lock(&vi_isr_lock);
	mfi_callback = NULL;
	mfi_callback_arg = NULL;
	mutex_unlock(&vi_isr_lock);

	return 0;
}
EXPORT_SYMBOL(tegra_vi_unregister_mfi_cb);
#endif

static void vi_mfi_worker(struct work_struct *vi_work)
{
	struct vi *pdev = container_of(vi_work, struct vi, mfi_cb_work);

	if (mfi_callback == NULL) {
		dev_dbg(&pdev->ndev->dev, "NULL callback\n");
		return;
	}
	mutex_lock(&vi_isr_lock);
	mfi_callback(mfi_callback_arg);
	mutex_unlock(&vi_isr_lock);
	return;
}

#if defined(CONFIG_TEGRA_ISOMGR)
static int vi_isomgr_unregister(struct vi *tegra_vi)
{
	tegra_isomgr_unregister(tegra_vi->isomgr_handle);
	tegra_vi->isomgr_handle = NULL;

	return 0;
}
#endif

static int vi_out_show(struct seq_file *s, void *unused)
{
	struct vi *vi = s->private;

	seq_printf(s, "vi overflow: %u\n",
		atomic_read(&(vi->vi_out.overflow)));

	return 0;
}

static int vi_out_open(struct inode *inode, struct file *file)
{
	return single_open(file, vi_out_show, inode->i_private);
}

static const struct file_operations vi_out_fops = {
	.open		= vi_out_open,
	.read		= seq_read,
	.llseek		= seq_lseek,
	.release	= single_release,
};

static void vi_remove_debugfs(struct vi *vi)
{
	debugfs_remove_recursive(vi->debugdir);
	vi->debugdir = NULL;
}

static void vi_create_debugfs(struct vi *vi)
{
	struct dentry *ret;
	char tegra_vi_name[20];
	char debugfs_file_name[20];


	snprintf(tegra_vi_name, sizeof(tegra_vi_name), "%s", TEGRA_VI_NAME);

	vi->debugdir = debugfs_create_dir(tegra_vi_name, NULL);
	if (!vi->debugdir) {
		dev_err(&vi->ndev->dev,
			"%s: failed to create %s directory",
			__func__, tegra_vi_name);
		goto create_debugfs_fail;
	}

	snprintf(debugfs_file_name, sizeof(debugfs_file_name), "%s", "vi_out");

	ret = debugfs_create_file(debugfs_file_name, S_IRUGO,
			vi->debugdir, vi, &vi_out_fops);
	if (!ret) {
		dev_err(&vi->ndev->dev,
		"%s: failed to create %s", __func__, debugfs_file_name);
		goto create_debugfs_fail;
	}

	return;

create_debugfs_fail:
	dev_err(&vi->ndev->dev, "%s: could not create debugfs", __func__);
	vi_remove_debugfs(vi);
}

static int nvhost_vi_slcg_handler(struct notifier_block *nb,
		unsigned long action, void *data)
{
	struct clk *clk;
	int ret = 0;

	struct nvhost_device_data *pdata =
		container_of(nb, struct nvhost_device_data,
			toggle_slcg_notifier);

	clk = clk_get(&pdata->pdev->dev, "pll_d");
	if (IS_ERR(clk))
		return -EINVAL;

	/* Make CSI sourced from PLL_D */
	ret = tegra_clk_cfg_ex(clk, TEGRA_CLK_PLLD_CSI_OUT_ENB, 1);
	if (ret) {
		dev_err(&pdata->pdev->dev,
		"%s: failed to select CSI source pll_d: %d\n",
		__func__, ret);
		return ret;
	}

	/* Enable PLL_D */
	ret = clk_prepare_enable(clk);
	if (ret) {
		dev_err(&pdata->pdev->dev, "Can't enable pll_d: %d\n", ret);
		return ret;
	}

	udelay(1);

	/* Disable PLL_D */
	clk_disable_unprepare(clk);

	/* Restore CSI source */
	ret = tegra_clk_cfg_ex(clk, TEGRA_CLK_MIPI_CSI_OUT_ENB, 1);
	if (ret) {
		dev_err(&pdata->pdev->dev,
		"%s: failed to restore csi source: %d\n",
		__func__, ret);
		return ret;
	}

	clk_put(clk);

	return NOTIFY_OK;
}

static int vi_probe(struct platform_device *dev)
{
	int err = 0;
	struct vi *tegra_vi;
	struct nvhost_device_data *pdata = NULL;

	if (dev->dev.of_node) {
		const struct of_device_id *match;

		match = of_match_device(tegra_vi_of_match, &dev->dev);
		if (match) {
			pdata = (struct nvhost_device_data *)match->data;
			dev->dev.platform_data = pdata;
		}
		/* DT initializes it to -1, use below WAR to set correct value.
		 * TODO: Once proper fix for dev-id goes in, remove it.
		 */
		dev->id = dev->dev.id;
	} else
		pdata = (struct nvhost_device_data *)dev->dev.platform_data;

	WARN_ON(!pdata);
	if (!pdata) {
		dev_info(&dev->dev, "no platform data\n");
		return -ENODATA;
	}

	err = nvhost_check_bondout(pdata->bond_out_id);
	if (err) {
		dev_warn(&dev->dev, "No VI unit present. err:%d", err);
		return err;
	}

	pdata->pdev = dev;
	mutex_init(&pdata->lock);
	platform_set_drvdata(dev, pdata);

	dev_info(&dev->dev, "%s: ++\n", __func__);

	tegra_vi = devm_kzalloc(&dev->dev, sizeof(struct vi), GFP_KERNEL);
	if (!tegra_vi) {
		dev_err(&dev->dev, "can't allocate memory for vi\n");
		return -ENOMEM;
	}

	err = nvhost_client_device_get_resources(dev);
	if (err)
		goto vi_probe_fail;

	tegra_vi->ndev = dev;
	tegra_vi->dev = &dev->dev;
	INIT_LIST_HEAD(&tegra_vi->entities);

	/* create workqueue for mfi callback */
	tegra_vi->vi_workqueue = alloc_workqueue("vi_workqueue",
					WQ_HIGHPRI | WQ_UNBOUND, 1);
	if (!tegra_vi->vi_workqueue) {
		dev_err(&dev->dev, "Failed to allocated vi_workqueue");
		goto vi_probe_fail;
	}
	/* Init mfi callback work */
	INIT_WORK(&tegra_vi->mfi_cb_work, vi_mfi_worker);

	/* call vi_intr_init and stats_work */
	INIT_WORK(&tegra_vi->stats_work, vi_stats_worker);

	err = vi_intr_init(tegra_vi);
	if (err)
		goto vi_probe_fail;

	vi_create_debugfs(tegra_vi);

	i2c_ctrl = pdata->private_data;
	pdata->private_data = tegra_vi;

	/* Create I2C Devices according to settings from board file */
	if (i2c_ctrl && i2c_ctrl->new_devices)
		i2c_ctrl->new_devices(dev);

	tegra_vi->reg = regulator_get(&tegra_vi->ndev->dev, "avdd_dsi_csi");
	if (IS_ERR(tegra_vi->reg)) {
		err = PTR_ERR(tegra_vi->reg);
		if (err == -ENODEV)
			dev_info(&tegra_vi->ndev->dev,
				"%s: no regulator device\n", __func__);
		else
			dev_err(&tegra_vi->ndev->dev,
				"%s: couldn't get regulator\n", __func__);
		tegra_vi->reg = NULL;
		if (tegra_platform_is_silicon())
			goto camera_i2c_unregister;
	}

#ifdef CONFIG_TEGRA_CAMERA
	tegra_vi->camera = tegra_camera_register(dev);
	if (!tegra_vi->camera) {
		dev_err(&dev->dev, "%s: can't register tegra_camera\n",
				__func__);
		goto vi_regulator_put;
	}
#endif

	if (pdata->slcg_notifier_enable &&
			(pdata->powergate_id != -1)) {
		pdata->toggle_slcg_notifier.notifier_call =
		&nvhost_vi_slcg_handler;

		slcg_register_notifier(pdata->powergate_id,
			&pdata->toggle_slcg_notifier);
	}

	nvhost_module_init(dev);

#ifdef CONFIG_PM_GENERIC_DOMAINS
#ifndef CONFIG_PM_GENERIC_DOMAINS_OF
	pdata->pd.name = "ve";
#endif

	/* add module power domain and also add its domain
	 * as sub-domain of MC domain */
	err = nvhost_module_add_domain(&pdata->pd, dev);
#endif

	err = nvhost_client_device_init(dev);
	if (err)
		goto camera_unregister;

	err = tegra_vi_v4l2_init(tegra_vi);
	if (err < 0)
		goto vi_init_error;

	/* Init Tegra VI channels */
	err = tegra_vi_channels_init(tegra_vi);
	if (err < 0)
		goto channels_error;

	/* Setup media links between VI and external sensor subdev. */
	err = tegra_vi_graph_init(tegra_vi);
	if (err < 0)
		goto graph_error;

	return 0;

graph_error:
	tegra_vi_channels_cleanup(tegra_vi);
channels_error:
	tegra_vi_v4l2_cleanup(tegra_vi);
vi_init_error:
	nvhost_client_device_release(dev);
camera_unregister:
#ifdef CONFIG_TEGRA_CAMERA
	tegra_camera_unregister(tegra_vi->camera);
vi_regulator_put:
#endif
	regulator_put(tegra_vi->reg);
	tegra_vi->reg = NULL;

camera_i2c_unregister:
	if (i2c_ctrl && i2c_ctrl->remove_devices)
		i2c_ctrl->remove_devices(dev);
	pdata->private_data = i2c_ctrl;
vi_probe_fail:
	dev_err(&dev->dev, "%s: failed\n", __func__);
	return err;
}

static int __exit vi_remove(struct platform_device *dev)
{
#ifdef CONFIG_TEGRA_CAMERA
	int err = 0;
#endif
	struct nvhost_device_data *pdata = platform_get_drvdata(dev);
	struct vi *tegra_vi = (struct vi *)pdata->private_data;

#ifdef CONFIG_PM_RUNTIME
	if (atomic_read(&dev->dev.power.usage_count) > 0)
		return -EBUSY;
#endif

	dev_info(&dev->dev, "%s: ++\n", __func__);

#if defined(CONFIG_TEGRA_ISOMGR)
	if (tegra_vi->isomgr_handle)
		vi_isomgr_unregister(tegra_vi);
#endif

	vi_remove_debugfs(tegra_vi);

	nvhost_client_device_release(dev);

	if (pdata->slcg_notifier_enable &&
	    (pdata->powergate_id != -1))
		slcg_unregister_notifier(pdata->powergate_id,
					 &pdata->toggle_slcg_notifier);

	vi_intr_free(tegra_vi);

	pdata->aperture[0] = NULL;
	flush_workqueue(tegra_vi->vi_workqueue);
	destroy_workqueue(tegra_vi->vi_workqueue);
#ifdef CONFIG_TEGRA_CAMERA
	err = tegra_camera_unregister(tegra_vi->camera);
	if (err)
		return err;
#endif

#ifdef CONFIG_PM_GENERIC_DOMAINS
	tegra_pd_remove_device(&dev->dev);
#endif

	regulator_put(tegra_vi->reg);
	tegra_vi->reg = NULL;

	/* Remove I2C Devices according to settings from board file */
	if (i2c_ctrl && i2c_ctrl->remove_devices)
		i2c_ctrl->remove_devices(dev);

	pdata->private_data = i2c_ctrl;

	tegra_vi_graph_cleanup(tegra_vi);
	tegra_vi_channels_cleanup(tegra_vi);
	tegra_vi_v4l2_cleanup(tegra_vi);

	return 0;
}

static struct platform_driver vi_driver = {
	.probe = vi_probe,
	.remove = __exit_p(vi_remove),
	.driver = {
		.owner = THIS_MODULE,
		.name = "vi",
#ifdef CONFIG_PM
		.pm = &nvhost_module_pm_ops,
#endif
#ifdef CONFIG_OF
		.of_match_table = tegra_vi_of_match,
#endif
	}
};

static struct of_device_id tegra_vi_domain_match[] = {
	{.compatible = "nvidia,tegra210-ve-pd",
	.data = (struct nvhost_device_data *)&t21_vi_info},
	{.compatible = "nvidia,tegra132-ve-pd",
	.data = (struct nvhost_device_data *)&t124_vi_info},
	{.compatible = "nvidia,tegra124-ve-pd",
	 .data = (struct nvhost_device_data *)&t124_vi_info},
	{},
};

static int __init vi_init(void)
{
	int ret;

	ret = nvhost_domain_init(tegra_vi_domain_match);
	if (ret)
		return ret;

	return platform_driver_register(&vi_driver);
}

static void __exit vi_exit(void)
{
	platform_driver_unregister(&vi_driver);
}

late_initcall(vi_init);
module_exit(vi_exit);
MODULE_LICENSE("GPL v2");
